@@ -5,6 +5,32 @@
 
 (in-package "SB-VM")
 
+;;; The SYNCHRONOUS-TRAP routine has nearly the same effect as executing INT3
+;;; but is more friendly to gdb. There may be some subtle bugs with regard to
+;;; blocking/unblocking of async signals which arrive nearly around the same
+;;; time as a synchronous trap.
+#+sw-int-avoidance ; "software interrupt avoidance"
+(define-assembly-routine (synchronous-trap) ()
+  (inst pushf)
+  (inst push rbp-tn)
+  (inst mov rbp-tn rsp-tn)
+  (inst and rsp-tn (- 16))
+  (inst sub rsp-tn 8) ; PUSHing an odd number of GPRs
+  ;; Arrange in the utterly confusing order that a linux signal context has them
+  ;; so that we can memcpy() into a context. Push RBX twice to maintain alignment.
+  (regs-pushlist rcx rax rdx rbx rbx rsi rdi r15 r14 r13 r12 r11 r10 r9 r8)
+  ;;                             ^^^ technically this is the slot for RBP
+  (inst sub rsp-tn (* 16 16))
+  (dotimes (i 16) (inst movdqa (ea (* i 16) rsp-tn) (sb-x86-64-asm::get-fpr :xmm i)))
+  (inst lea rdi-tn (ea 24 rbp-tn)) ; stack-pointer at moment of "interrupt"
+  (inst mov rsi-tn rsp-tn)         ; pointer to saved CPU state
+  (inst call (make-fixup "synchronous_trap" :foreign))
+  (dotimes (i 16) (inst movdqa (sb-x86-64-asm::get-fpr :xmm i) (ea (* i 16) rsp-tn)))
+  (inst add rsp-tn (* 16 16))
+  (regs-poplist rcx rax rdx rbx rbx rsi rdi r15 r14 r13 r12 r11 r10 r9 r8)
+  (inst leave)
+  (inst popf))
+
 (macrolet ((do-fprs (operation regset &aux (displacement 0))
              ;; The YMM case could be removed now I suppose, since we use XSAVE + XRSTOR
              (multiple-value-bind (mnemonic fpr-align)
@@ -58,12 +84,11 @@
     (inst pop rdx-tn)))
 
 (define-assembly-routine (switch-to-arena (:return-style :raw)) ()
-  (inst mov rsi-tn (ea rsp-tn)) ; explicitly  pass the return PC
   ;; RSI and RDI are vop temps, so don't bother preserving them
   (with-registers-preserved (c :except (rsi rdi))
     (pseudo-atomic ()
       #-system-tlabs (inst break halt-trap)
-      #+system-tlabs (inst call (make-fixup "switch_to_arena" :foreign)))))
+      #+system-tlabs (call-c (make-fixup "switch_to_arena" :foreign) #+win32 rdi-tn #+win32 rsi-tn))))
 
 (macrolet ((def-routine-pair (name&options vars &body code)
              `(progn
@@ -79,54 +104,58 @@
 
 (def-routine-pair (alloc-tramp) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst mov rsi-tn system-tlab-p)
-    (inst call (make-fixup "alloc" :foreign))
+    (call-c (make-fixup "alloc" :foreign)
+            (ea 16 rbp-tn)
+            system-tlab-p)
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 (def-routine-pair (list-alloc-tramp) () ; CONS, ACONS, LIST, LIST*
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst mov rsi-tn system-tlab-p)
-    (inst call (make-fixup "alloc_list" :foreign))
+    (call-c (make-fixup "alloc_list" :foreign)
+            (ea 16 rbp-tn)
+            system-tlab-p)
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 (def-routine-pair (listify-&rest (:return-style :none)) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn)) ; 1st C call arg
-    (inst mov rsi-tn (ea 24 rbp-tn)) ; 2nd C call arg
-    (inst mov rdx-tn system-tlab-p)
-    (inst call (make-fixup "listify_rest_arg" :foreign))
-    (inst mov (ea 24 rbp-tn) rax-tn)) ; result
+    (call-c (make-fixup "listify_rest_arg" :foreign)
+            (ea 16 rbp-tn)
+            (ea 24 rbp-tn)
+            system-tlab-p)
+    (inst mov (ea 24 rbp-tn) rax-tn))   ; result
   (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 
 (def-routine-pair (make-list (:return-style :none)) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn)) ; 1st C call arg
-    (inst mov rsi-tn (ea 24 rbp-tn)) ; 2nd C call arg
-    (inst mov rdx-tn system-tlab-p)
-    (inst call (make-fixup "make_list" :foreign))
+    (call-c (make-fixup "make_list" :foreign)
+            (ea 16 rbp-tn)
+            (ea 24 rbp-tn)
+            system-tlab-p)
     (inst mov (ea 24 rbp-tn) rax-tn)) ; result
   (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 )
 
 (define-assembly-routine (alloc-funinstance) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup "alloc_funinstance" :foreign))
+    (call-c (make-fixup "alloc_funinstance" :foreign)
+            (ea 16 rbp-tn))
     (inst mov (ea 16 rbp-tn) rax-tn)))
 
 ;;; These routines are for the deterministic consing profiler.
 ;;; The C support routine's argument is the return PC.
 (define-assembly-routine (enable-alloc-counter) ()
   (with-registers-preserved (c)
-    (inst lea rdi-tn (ea 8 rbp-tn))
-    (pseudo-atomic () (inst call (make-fixup "allocation_tracker_counted" :foreign)))))
+    #+sb-thread
+    (pseudo-atomic ()
+      (call-c (make-fixup "allocation_tracker_counted" :foreign)
+              (* (ea 8 rbp-tn))))))
 
 (define-assembly-routine (enable-sized-alloc-counter) ()
   (with-registers-preserved (c)
-    (inst lea rdi-tn (ea 8 rbp-tn))
-    (pseudo-atomic () (inst call (make-fixup "allocation_tracker_sized" :foreign)))))
+    #+sb-thread
+    (pseudo-atomic ()
+      (call-c (make-fixup "allocation_tracker_sized" :foreign)
+              (* (ea 8 rbp-tn))))))
 
 (define-assembly-routine (undefined-tramp (:return-style :none))
     ((:temp rax descriptor-reg rax-offset))
@@ -211,3 +240,56 @@
     (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
     (call-static-fun 'ensure-symbol-hash 1)
     (inst mov (ea 16 rbp-tn) rdx-tn))) ; result to arg passing loc
+
+;;; Perform a store to code, updating the GC card mark bit.
+;;; This has two additional complications beyond the ordinary
+;;; generational barrier:
+;;; 1. immobile code uses its own card table which maps linearly
+;;;    with the page index, unlike the dynamic space card table
+;;;    that has a different way of computing a card address.
+;;; 2. code objects are so seldom written that it behooves us to
+;;;    track within each object whether it has been written,
+;;;    thereby avoiding scanning of unwritten objects.
+;;;    This is especially important for immobile space where
+;;;    it is likely that new code will be co-located on a page
+;;;    with old code due to the non-moving allocator.
+(define-assembly-routine (code-header-set (:return-style :none)) ()
+  ;; stack: ret-pc, object, index, value-to-store
+  (symbol-macrolet ((object (ea 8 rsp-tn))
+                    (word-index (ea 16 rsp-tn))
+                    (newval (ea 24 rsp-tn))
+                    ;; these are declared as vop temporaries
+                    (rax rax-tn)
+                    (rdx rdx-tn)
+                    (rdi rdi-tn))
+    (pseudo-atomic ()
+      #+immobile-space
+      (progn
+        #-sb-thread
+        (let ((fixup (make-fixup "all_threads" :foreign-dataref)))
+          ;; Load THREAD-BASE-TN from the all_threads. Does not need to be spilled
+          ;; to stack, because we do do not give the register allocator access to it.
+          (inst mov thread-tn (rip-relative-ea fixup))
+          (inst mov thread-tn (ea thread-tn)))
+        (inst mov rax object)
+        (inst sub rax (thread-slot-ea thread-text-space-addr-slot))
+        (inst shr rax (1- (integer-length immobile-card-bytes)))
+        (inst cmp rax (thread-slot-ea thread-text-card-count-slot))
+        (inst jmp :ae try-dynamic-space)
+        (inst mov rdi (thread-slot-ea thread-text-card-marks-slot))
+        (inst bts :dword :lock (ea rdi-tn) rax)
+        (inst jmp store))
+      TRY-DYNAMIC-SPACE
+      (inst mov rax object)
+      (inst shr rax gencgc-card-shift)
+      (inst and :dword rax card-index-mask)
+      (inst mov :byte (ea gc-card-table-reg-tn rax) CARD-MARKED)
+      STORE
+      (inst mov rdi object)
+      (inst mov rdx word-index)
+      (inst mov rax newval)
+      ;; set 'written' flag in the code header
+      (inst or :byte :lock (ea (- 3 other-pointer-lowtag) rdi) #x40)
+      ;; store newval into object
+      (inst mov (ea (- other-pointer-lowtag) rdi rdx n-word-bytes) rax)))
+  (inst ret 24)) ; remove 3 stack args
