@@ -25,7 +25,7 @@
 
 (defun guess-alignment (bits)
   (cond ((null bits) nil)
-        #-(and x86 (not win32)) ((> bits 32) 64)
+        #-(or (and x86 (not win32)) (and ppc darwin)) ((> bits 32) 64)
         ((> bits 16) 32)
         ((> bits 8) 16)
         ((> bits 1) 8)
@@ -326,6 +326,7 @@
   "Define the alien type NAME to be equivalent to TYPE. Name may be NIL for
    STRUCT and UNION types, in which case the name is taken from the type
    specifier."
+  (declare (symbol name))
   (with-auxiliary-alien-types env
     (let ((alien-type (parse-alien-type type env)))
       `(eval-when (:compile-toplevel :load-toplevel :execute)
@@ -1431,7 +1432,7 @@
          ;; Only pin things on GENCGC, since on CHENEYGC it'd imply
          ;; disabling the GC.  Which is something we don't want to do
          ;; every time we're calling to C.
-         #+gencgc
+         #-cheneygc
          (loop for variable in variables
             for type in types
             when (invoke-alien-type-method :deport-pin-p type)
