@@ -74,7 +74,7 @@
   ;; Specifically, hashes that depend on object identity (address) are impermissible.
   (dolist (piece pieces)
     (aver (typep piece '(or string symbol (cons (eql :character-set) string)))))
-  (%make-pattern (sxhash pieces) pieces))
+  (%make-pattern (pathname-sxhash pieces) pieces))
 
 (declaim (inline %pathname-directory))
 (defun %pathname-directory (pathname) (car (%pathname-dir+hash pathname)))
@@ -290,6 +290,10 @@
   (or (eq (car entry) (car key)) ; quick win if lists are EQ
       (and (eq (cdr entry) (cdr key)) ; hashes match
            (compare-component (car entry) (car key)))))
+(defun pn-table-hash (pathname)
+  ;; The pathname table makes distinctions between pathnames that EQUAL does not.
+  (mix (sxhash (%pathname-version pathname))
+       (pathname-sxhash pathname)))
 (defun pn-table-pn= (entry key)
   (and (compare-pathname-host/dev/dir/name/type entry key)
        (eql (%pathname-version entry) (%pathname-version key))))
@@ -297,7 +301,7 @@
 (defun !pathname-cold-init ()
   (setq *pn-dir-table* (make-hashset 32 #'pn-table-dir= #'cdr
                                      :synchronized t :weakness t)
-        *pn-table* (make-hashset 32 #'pn-table-pn= #'pathname-sxhash
+        *pn-table* (make-hashset 32 #'pn-table-pn= #'pn-table-hash
                                  :synchronized t :weakness t)))
 
 ;;; A pathname is logical if the host component is a logical host.
@@ -321,9 +325,11 @@
     (flet ((ensure-heap-string (part) ; return any non-string as-is
              ;; FIXME: what about pattern pieces and (:HOME "user") ?
              (cond ((or (not (stringp part)) (read-only-space-obj-p part)) part)
-                   ((dynamic-space-obj-p part) (logically-readonlyize part))
-                   ;; dynamic-extent strings and lisp strings in alien memory are acceptable.
-                   ;; Copies must be made in that case, since we're holding on to them.
+                   ;; Altering a string that is any piece of any arg to INTERN-PATHNAME
+                   ;; is a surefire way to corrupt the hashset, so *always* copy the input.
+                   ;; Users might not have reason to believe that once a string is passed
+                   ;; to any pathname function, it is immutable. We can only hope that
+                   ;; they don't mutate strings returned by pathname accessors.
                    (t (let ((l (length part)))
                         (logically-readonlyize
                          (replace (typecase part
@@ -337,7 +343,7 @@
                    (lambda (dir)
                      (cons (mapcar #'ensure-heap-string (car dir)) (cdr dir))))))
              (pn-key (!allocate-pathname host device dir+hash name type version)))
-        (declare (truly-dynamic-extent pn-key))
+        (declare (dynamic-extent pn-key))
         (hashset-insert-if-absent
          *pn-table* pn-key
          (lambda (tmp &aux (host (%pathname-host tmp)))
@@ -360,36 +366,39 @@
   (hashset-rehash *pn-table* nil))
 
 (defun show-pn-cache (&aux (*print-pretty* nil) (*package* (find-package "CL-USER")))
-  (without-gcing
-   (dolist (symbol '(*pn-dir-table* *pn-table*))
-     (let* ((hashset (symbol-value symbol))
-            (dirs (hss-cells (hashset-storage *pn-dir-table*)))
-            (v (hss-cells (hashset-storage hashset)))
-            (n (hs-cells-capacity v)))
-       (format t "~&~S: size=~D tombstones=~D unused=~D~%" symbol n
-               (count nil v :end n) (count 0 v :end n))
-       (dotimes (i n)
-         (let ((entry (aref v i)))
-           (unless (member entry '(nil 0))
-             (if (eq symbol '*pn-dir-table*)
-                 (format t "~4d ~3d ~x ~16x ~s~%" i
-                         (generation-of entry) (get-lisp-obj-address entry)
-                         (cdr entry) (car entry))
-                 (format t "~4d ~3d ~x ~16x [~A ~S ~A ~S ~S ~S]~%"
-                         i (generation-of entry) (get-lisp-obj-address entry)
-                         (pathname-sxhash entry)
-                         (let ((host (%pathname-host entry)))
-                           (cond ((logical-host-p host)
-                                  (prin1-to-string (logical-host-name host)))
-                                 ((eq host *physical-host*) "phys")
-                                 (t host)))
-                         (%pathname-device entry)
-                         (acond ((%pathname-dir+hash entry)
-                                 (format nil "@~D" (position it dirs)))
-                                (t "-"))
-                         (%pathname-name entry)
-                         (%pathname-type entry)
-                         (%pathname-version entry))))))))))
+  (dolist (symbol '(*pn-dir-table* *pn-table*))
+    (let* ((hashset (symbol-value symbol))
+           (v (hss-cells (hashset-storage hashset)))
+           (n (hs-cells-capacity v)))
+      (multiple-value-bind (live tombstones unused) (hs-cells-occupancy v n)
+        (declare (ignore live))
+        (format t "~&~S: size=~D tombstones=~D unused=~D~%" symbol n tombstones unused))
+      (dotimes (i n)
+        (let ((entry (hs-cell-ref v i)))
+          (unless (member entry '(nil 0))
+            (format t "~4d ~3d ~x " i (generation-of entry) (get-lisp-obj-address entry))
+            (if (eq symbol '*pn-dir-table*)
+                (format t "~16x ~s~%" (cdr entry) (car entry))
+                (flet ((index-of (pn-dir)
+                         (let ((dirs (hss-cells (hashset-storage *pn-dir-table*))))
+                           (dotimes (i (weak-vector-len dirs))
+                             (when (eq pn-dir (weak-vector-ref dirs i))
+                               (return i))))))
+                  (format t
+                   "~16x [~A ~S ~A ~S ~S ~S]~%"
+                   (pathname-sxhash entry)
+                   (let ((host (%pathname-host entry)))
+                     (cond ((logical-host-p host)
+                            ;; display with string quotes around name
+                            (prin1-to-string (logical-host-name host)))
+                           ((eq host *physical-host*) "phys")
+                           (t host)))
+                   (%pathname-device entry)
+                   (acond ((%pathname-dir+hash entry) (format nil "@~D" (index-of it)))
+                          (t "-"))
+                   (%pathname-name entry)
+                   (%pathname-type entry)
+                   (%pathname-version entry))))))))))
 
 ;;; Vector of logical host objects, each of which contains its translations.
 ;;; The vector is never mutated- always a new vector is created when adding
@@ -557,8 +566,6 @@
 (defun pathname= (a b)
   (declare (type pathname a b))
   (or (eq a b)
-      ;; I believe that this is actually the same as EQ on pathnames now,
-      ;; but until I prove it, I'm leaving this code in.
       (and (compare-pathname-host/dev/dir/name/type a b)
            (or (eq (%pathname-host a) *physical-host*)
                (compare-component (pathname-version a)
@@ -567,18 +574,27 @@
 (sb-kernel::assign-equalp-impl 'pathname #'pathname=)
 (sb-kernel::assign-equalp-impl 'logical-pathname #'pathname=)
 
-;;; Hash either a PATHNAME or a PATHNAME-DIRECTORY. This is called byt both SXHASH
-;;; and by the interning of pathnames, which uses a multi-step approaching to
-;;; coalescing shared subparts. If an EQUAL directory was used before, we share that.
+;;; Hash a PATHNAME or a PATHNAME-DIRECTORY or pieces of a PATTERN.
+;;; This is called by both SXHASH and by the interning of pathnames, which uses a
+;;; multi-step approaching to coalescing shared subparts.
+;;; If an EQUAL directory was used before, we share that.
 ;;; Since a directory is stored with its hash precomputed, hashing a PATHNAME as a
 ;;; whole entails at most 4 more MIX operations. So using pathnames as keys in
 ;;; a hash-table pays a small up-front price for later speed improvement.
 (defun pathname-sxhash (x)
-  (flet ((hash-piece (piece)
+  (labels
+      ((hash-piece (piece)
            (etypecase piece
-             (string (sxhash piece)) ; transformed
+             (string
+              (let ((res (length piece)))
+                (if (<= res 6) ; hash it more thoroughly than (SXHASH string)
+                    (dovector (ch piece res)
+                      (setf res (mix (murmur-hash-word/+fixnum (char-code ch)) res)))
+                    (sxhash piece))))
              (symbol (sxhash piece)) ; transformed
              (pattern (pattern-hash piece))
+             ;; next case is only for MAKE-PATTERN
+             ((cons (eql :character-set)) (hash-piece (the string (cdr piece))))
              ((cons (eql :home) (cons string null))
               ;; :HOME has two representations- one is just '(:absolute :home ...)
               ;; and the other '(:absolute (:home "user") ...)
@@ -592,10 +608,13 @@
          (awhen (%pathname-dir+hash x) (mixf hash (cdr it)))
          (mixf hash (hash-piece (%pathname-name x)))
          (mixf hash (hash-piece (%pathname-type x)))
-         ;; EQUAL might ignore the version, and it doesn't provide many bits
-         ;; of randomness, so don't bother with it.
+         ;; The requirement NOT to mix the version into the resulting hash is mandated
+         ;; by bullet point 1 in the SXHASH specification:
+         ;;  (equal x y) implies (= (sxhash x) (sxhash y))
+         ;; and the observation that in this implementation of Lisp:
+         ;;  (equal (make-pathname :version 1) (make-pathname :version 15)) => T
          hash))
-      (list ;; a directory
+      (list ;; a directory, or the PIECES argument to MAKE-PATTERN
        (let ((hash 0))
          (dolist (piece x hash)
            (mixf hash (hash-piece piece))))))))
