@@ -26,9 +26,15 @@
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg zero)))
   (:info name offset lowtag)
-  (:ignore name)
   (:results)
+  (:vop-var vop)
+  (:node-var node)
   (:generator 1
+    (unless (or
+             ;; gencgc does not need to emit the barrier for constructors
+             (eq name :allocator)
+             (sb-c::set-slot-old-p node))
+      (emit-gengc-barrier object nil tmp-tn (vop-nth-arg 1 vop) value))
     (storew value object offset lowtag)))
 
 (define-vop (compare-and-swap-slot)
@@ -39,7 +45,9 @@
   (:ignore name)
   (:temporary (:sc non-descriptor-reg) lip)
   (:results (result :scs (descriptor-reg any-reg) :from :load))
+  (:vop-var vop)
   (:generator 5
+    (emit-gengc-barrier object nil lip (vop-nth-arg 2 vop) new)
     (inst add-sub lip object (- (* offset n-word-bytes) lowtag))
     (cond ((member :arm-v8.1 *backend-subfeatures*)
            (move result old)
@@ -81,6 +89,7 @@
                                       (symbol-always-has-tls-value-p (tn-value object)))))
            (value :scs (descriptor-reg any-reg zero)))
     (:temporary (:sc any-reg) tls-index)
+    (:vop-var vop)
     (:generator 4
       (sc-case object
         (constant
@@ -91,6 +100,7 @@
            (inst ldr tmp-tn (@ thread-tn tls-index))
            (compare-to-no-tls-value-marker tmp-tn)
            (inst b :ne LOCAL)
+           (emit-gengc-barrier object nil tls-index (vop-nth-arg 1 vop) value)
            (storew value object symbol-value-slot other-pointer-lowtag)
            (inst b DONE)
            LOCAL
@@ -244,6 +254,9 @@
       (loadw tmp-tn symbol symbol-name-slot other-pointer-lowtag)
       (inst and result tmp-tn (1- (ash 1 sb-impl::symbol-name-bits))))))
 
+;; Is it worth eliding the GC store barrier for a thread-local CAS?
+;; Not really, because why does such an operation even exists?
+;; No other thread can see our TLS except via SAP-REF.
 (define-vop (%compare-and-swap-symbol-value)
   (:translate %compare-and-swap-symbol-value)
   (:args (symbol :scs (descriptor-reg))
@@ -256,6 +269,7 @@
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 15
+    (emit-gengc-barrier symbol nil lip (vop-nth-arg 2 vop) new)
     (inst dsb)
     #+sb-thread
     (assemble ()
@@ -299,6 +313,7 @@
   (:vop-var vop)
   (:guard (member :arm-v8.1 *backend-subfeatures*))
   (:generator 14
+    (emit-gengc-barrier symbol nil lip (vop-nth-arg 2 vop) new)
     #+sb-thread
     (assemble ()
       (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
@@ -346,6 +361,7 @@
   (:temporary (:scs (non-descriptor-reg)) type)
   (:results (result :scs (descriptor-reg)))
   (:generator 38
+    (emit-gengc-barrier fdefn nil lip)
     (inst add-sub lip function (- (* simple-fun-insts-offset n-word-bytes)
                                   fun-pointer-lowtag))
     (load-type type function (- fun-pointer-lowtag))
@@ -583,6 +599,48 @@
 (define-full-setter instance-index-set * instance-slots-offset
   instance-pointer-lowtag (descriptor-reg any-reg) * %instance-set)
 
+(define-vop (instance-set-multiple)
+  (:args (instance :scs (descriptor-reg))
+         (values :more t :scs (descriptor-reg any-reg)))
+  (:temporary (:sc descriptor-reg) val-temp)
+  (:info indices)
+  (:generator 1
+    (emit-gengc-barrier instance nil tmp-tn values)
+    (do ((tn-ref values (tn-ref-across tn-ref))
+         (indices indices (cdr indices)))
+        ((null tn-ref))
+      (let* ((value (tn-ref-tn tn-ref))
+             (source
+              (sc-case value
+               (immediate
+                ;; Pretty sure this should not be a copy-and-paste from
+                ;; LOAD-IMMEDIATE. What am I doing wrong?
+                (let ((val (tn-value value)))
+                  (etypecase val
+                    (integer
+                     ;; This is a FIXNUM, as IMMEDIATE-CONSTANT-SC only
+                     ;; accepts integers if they are FIXNUMs.
+                     (load-immediate-word val-temp (fixnumize val)))
+                    (character
+                     (let* ((codepoint (char-code val))
+                            (tagged (dpb codepoint (byte 24 8) character-widetag)))
+                       (load-immediate-word val-temp tagged)))
+                    (symbol
+                     (load-symbol val-temp val))))
+                val-temp)
+               (constant
+                (inst load-constant val-temp (tn-byte-offset value))
+                val-temp)
+               (control-stack
+                (load-stack-tn val-temp value)
+                val-temp)
+               ((any-reg descriptor-reg)
+                value))))
+        (inst str source
+              (@ instance (load-store-offset
+                           (- (ash (+ instance-slots-offset (car indices)) word-shift)
+                              instance-pointer-lowtag))))))))
+
 (define-vop (%instance-cas word-index-cas)
   (:policy :fast-safe)
   (:translate %instance-cas)
@@ -650,8 +708,6 @@
   (:temporary (:scs (non-descriptor-reg)) temp card)
   (:temporary (:sc non-descriptor-reg) pa-flag)
   (:generator 10
-    (load-foreign-symbol temp "gc_card_table_mask" :dataref t)
-    (inst ldr (32-bit-reg temp) (@ temp)) ; 4-byte int
     (pseudo-atomic (pa-flag)
       #+immobile-space
       (progn
@@ -686,6 +742,8 @@
                  (inst cbnz (32-bit-reg pa-flag) LOOP))))
         (inst b store))
       TRY-DYNAMIC-SPACE
+      (load-foreign-symbol temp "gc_card_table_mask" :dataref t)
+      (inst ldr (32-bit-reg temp) (@ temp)) ; 4-byte int
       ;; Compute card mark index
       (inst lsr card object gencgc-card-shift)
       (inst and card card temp)
