@@ -10,20 +10,23 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;; Return the maximum number of bits an integer of the supplied type
 ;;; can take up, or NIL if it is unbounded. The second (third) value
 ;;; is T if the integer can be positive (negative) and NIL if not.
 ;;; Zero counts as positive.
 (defun integer-type-length (type)
-  (if (numeric-type-p type)
+  (if (and (numeric-type-p type)
+           (eq (numeric-type-class type) 'integer))
       (let ((min (numeric-type-low type))
             (max (numeric-type-high type)))
         (values (and min max (max (integer-length min) (integer-length max)))
                 (or (null max) (not (minusp max)))
-                (or (null min) (minusp min))))
-      (values nil t t)))
+                (or (null min) (minusp min))
+                min
+                max))
+      (values nil t t nil nil)))
 
 ;;;; Generators for simple bit masks
 
@@ -138,10 +141,14 @@
 (defun logand-derive-type-aux (x y &optional same-leaf)
   (when same-leaf
     (return-from logand-derive-type-aux x))
-  (multiple-value-bind (x-len x-pos x-neg) (integer-type-length x)
-    (declare (ignore x-pos))
-    (multiple-value-bind (y-len y-pos y-neg) (integer-type-length y)
-      (declare (ignore y-pos))
+  (flet ((minus-one (x y)
+           (when (eql (nth-value 1 (type-singleton-p x))
+                      -1)
+             (return-from logand-derive-type-aux y))))
+    (minus-one x y)
+    (minus-one y x))
+  (multiple-value-bind (x-len x-pos x-neg x-low x-high) (integer-type-length x)
+    (multiple-value-bind (y-len y-pos y-neg y-low y-high) (integer-type-length y)
       (if (not x-neg)
           ;; X must be positive.
           (if (not y-neg)
@@ -157,55 +164,50 @@
                          (logand-derive-unsigned-bounds x y)
                        (specifier-type `(integer ,low ,high)))))
               ;; X is positive, but Y might be negative.
-              (cond ((null x-len)
-                     (specifier-type 'unsigned-byte))
+              (cond ((and x-len y-low y-high (< y-high 0))
+                     (multiple-value-bind (low high)
+                         (let ((len (max x-len y-len)))
+                           (logand-derive-unsigned-bounds x (make-numeric-type :class 'integer
+                                                                               :low (ldb (byte len 0) y-low)
+                                                                               :high (ldb (byte len 0) y-high))))
+                       (specifier-type `(integer ,low ,high))))
+                    (x-high
+                     (specifier-type `(integer 0 ,x-high)))
                     (t
-                     (specifier-type `(unsigned-byte* ,x-len)))))
+                     (specifier-type 'unsigned-byte))))
           ;; X might be negative.
           (if (not y-neg)
               ;; Y must be positive.
-              (cond ((null y-len)
-                     (specifier-type 'unsigned-byte))
-                    (t (specifier-type `(unsigned-byte* ,y-len))))
+              (cond ((and y-len x-low x-high (< x-high 0))
+                     (multiple-value-bind (low high)
+                         (let ((len (max x-len y-len)))
+                           (logand-derive-unsigned-bounds y (make-numeric-type :class 'integer
+                                                                               :low (ldb (byte len 0) x-low)
+                                                                               :high (ldb (byte len 0) x-high))))
+                       (specifier-type `(integer ,low ,high))))
+                    (y-high
+                     (specifier-type `(integer 0 ,y-high)))
+                    (t
+                     (specifier-type 'unsigned-byte)))
               ;; Either might be negative.
-              (if (and x-len y-len)
-                  ;; The result is bounded.
-                  (specifier-type `(signed-byte ,(1+ (max x-len y-len))))
-                  ;; We can't tell squat about the result.
-                  (specifier-type 'integer)))))))
-
-(defun make-modular-fun-type-deriver (prototype kind width signedp)
-  (declare (ignore kind))
-  #!-sb-fluid
-  (binding* ((info (info :function :info prototype) :exit-if-null)
-             (fun (fun-info-derive-type info) :exit-if-null)
-             (mask-type (specifier-type
-                         (ecase signedp
-                             ((nil) (let ((mask (1- (ash 1 width))))
-                                      `(integer ,mask ,mask)))
-                             ((t) `(signed-byte ,width))))))
-    (lambda (call)
-      (let ((res (funcall fun call)))
-        (when res
-          (if (eq signedp nil)
-              (logand-derive-type-aux res mask-type))))))
-  #!+sb-fluid
-  (lambda (call)
-    (binding* ((info (info :function :info prototype) :exit-if-null)
-               (fun (fun-info-derive-type info) :exit-if-null)
-               (res (funcall fun call) :exit-if-null)
-               (mask-type (specifier-type
-                           (ecase signedp
-                             ((nil) (let ((mask (1- (ash 1 width))))
-                                      `(integer ,mask ,mask)))
-                             ((t) `(signed-byte ,width))))))
-      (if (eq signedp nil)
-          (logand-derive-type-aux res mask-type)))))
+              (cond ((and x-low y-low)
+                     (specifier-type `(integer ,(zeroes (integer-length (min x-low y-low)))
+                                               ,(if (and x-high y-high)
+                                                    (max x-high y-high -1)
+                                                    '*))))
+                    ((and x-high y-high)
+                     (specifier-type `(integer *
+                                               ,(max x-high y-high -1))))
+                    (t
+                     (if (and (not y-pos)
+                              (not x-pos))
+                         (specifier-type '(integer * -1))
+                         (specifier-type 'integer)))))))))
 
 (defun logior-derive-unsigned-bounds (x y)
-  (let* ((a (numeric-type-low x))
+  (let* ((a (max (or (numeric-type-low x) 0) 0))
          (b (numeric-type-high x))
-         (c (numeric-type-low y))
+         (c (max (or (numeric-type-low y) 0) 0))
          (d (numeric-type-high y))
          (length-xor-x (integer-length (logxor a b)))
          (length-xor-y (integer-length (logxor c d))))
@@ -228,8 +230,8 @@
 (defun logior-derive-type-aux (x y &optional same-leaf)
   (when same-leaf
     (return-from logior-derive-type-aux x))
-  (multiple-value-bind (x-len x-pos x-neg) (integer-type-length x)
-    (multiple-value-bind (y-len y-pos y-neg) (integer-type-length y)
+  (multiple-value-bind (x-len x-pos x-neg x-low x-high) (integer-type-length x)
+    (multiple-value-bind (y-len y-pos y-neg y-low y-high) (integer-type-length y)
       (cond
        ((and (not x-neg) (not y-neg))
         ;; Both are positive.
@@ -251,21 +253,35 @@
             ;; X is negative, but we don't know about Y. The result
             ;; will be negative, but no more negative than X.
             (specifier-type
-             `(integer ,(or (numeric-type-low x) '*)
+             `(integer ,(or x-low '*)
                        -1))))
        (t
         ;; X might be either positive or negative.
         (if (not y-pos)
             ;; But Y is negative. The result will be negative.
             (specifier-type
-             `(integer ,(or (numeric-type-low y) '*)
+             `(integer ,(or y-low '*)
                        -1))
-            ;; We don't know squat about either. It won't get any bigger.
-            (if (and x-len y-len)
-                ;; Bounded.
-                (specifier-type `(signed-byte ,(1+ (max x-len y-len))))
-                ;; Unbounded.
-                (specifier-type 'integer))))))))
+            (let ((type
+                    (cond ((and x-len y-len)
+                           (specifier-type `(integer ,(min x-low y-low)
+                                                     ,(nth-value 1 (logior-derive-unsigned-bounds x y)))))
+                          ((and x-high y-high)
+                           (specifier-type `(integer * ,(nth-value 1 (logior-derive-unsigned-bounds x y)))))
+                          (t
+                           (specifier-type 'integer)))))
+              (flet ((contains-zero (low high)
+                       (cond ((and low high)
+                              (<= low 0 high))
+                             (low
+                              (<= low 0))
+                             (high
+                              (>= high 0))
+                             (t))))
+               (if (and (contains-zero x-low x-high)
+                        (contains-zero y-low y-high))
+                   type
+                   (type-intersection type (specifier-type '(and integer (not (eql 0))))))))))))))
 
 (defun logxor-derive-unsigned-bounds (x y)
   (let* ((a (numeric-type-low x))
