@@ -222,7 +222,8 @@
 ;;; EMIT-MOVES-AND-COERCIONS. That's wasteful.
 (defun make-constant-tn (constant &optional force-boxed)
   (declare (type constant constant))
-  (or (leaf-info constant)
+  (or (and (tn-p (leaf-info constant))
+           (leaf-info constant))
       (multiple-value-bind (immed null-offset)
           (immediate-constant-sc (constant-value constant))
         ;; currently NULL-OFFSET is used only on ARM64
@@ -375,11 +376,13 @@
   (when refs
     (let* ((first (reference-tn (tn-ref-tn refs) write-p))
            (prev first))
+      (setf (tn-ref-type first) (tn-ref-type refs))
       (loop for tn-ref = (tn-ref-across refs) then (tn-ref-across tn-ref)
             while tn-ref
             do
-            (let ((ref (reference-tn  (tn-ref-tn tn-ref) write-p)))
-              (setf (tn-ref-across prev) ref)
+            (let ((ref (reference-tn (tn-ref-tn tn-ref) write-p)))
+              (setf (tn-ref-across prev) ref
+                    (tn-ref-type ref) (tn-ref-type tn-ref))
               (setq prev ref)))
       first)))
 
