@@ -50,6 +50,15 @@
              (inst mov :qword (object-slot-ea object 1 lowtag) nil-value))
            (inst mov :dword (vector-len-ea object)
                  (or (encode-value-if-immediate value) value)))
+          #-soft-card-marks
+          ((or (equal name '(setf %funcallable-instance-fun)) (eq name '%set-fun-layout))
+           ;; If soft card marks are disabled, then EMIT-GENGC-BARRIER is disabled too.
+           ;; But funcallable-instances are on PAGE_TYPE_CODE, and code pages do not use
+           ;; MMU-based protection regardless of this feature.
+           ;; So we have to alter the card mark differently.
+           (pseudo-atomic ()
+             (emit-code-page-gengc-barrier object val-temp)
+             (emit-store (object-slot-ea object offset lowtag) value val-temp)))
           (t
            (let* ((value-tn (tn-ref-tn (tn-ref-across args)))
                   (prim-type (sb-c::tn-primitive-type value-tn))
@@ -156,30 +165,28 @@
              (eq (info :variable :wired-tls (tn-value symbol)) :always-thread-local))
         ;; We never need the GC barrier for TLS.  I think it would be preferable
         ;; to resolve this in IR1, maybe turning it into SET-TLS-VALUE.
-        (emit-store (ea (make-fixup (tn-value symbol) :symbol-tls-index) thread-tn)
-                      value val-temp)
-        (let ((store (gen-label)))
-          (cond ((and (sc-is symbol immediate)
-                      (info :variable :wired-tls (tn-value symbol)))
-                 ;; The TLS index is arbitrary but known to be nonzero,
-                 ;; so we can resolve the displacement of the thread-local value
-                 ;; at load-time, saving one instruction over the general case.
-                 (inst lea cell (ea (make-fixup (tn-value symbol) :symbol-tls-index)
-                                    thread-tn)))
-                (t
-                 ;; These MOVs look the same, but when the symbol is immediate, this is
-                 ;; a load from an absolute address. Needless to say, the names of these
-                 ;; accessor macros are arbitrary - the difference is not very apparent.
-                 (inst mov :dword cell (if (sc-is symbol immediate)
-                                           (symbol-tls-index-ea symbol)
-                                           (tls-index-of symbol)))
-                 (inst add cell thread-tn)))
-          (inst cmp :qword (ea cell) no-tls-value-marker)
-          (inst jmp :ne STORE)
-          (emit-symbol-write-barrier symbol nil val-temp (vop-nth-arg 1 vop) value)
-          (get-symbol-value-slot-ea cell symbol)
-          (emit-label STORE)
-          (emit-store (ea cell) value val-temp)))))
+        (return-from set
+          (emit-store (ea (make-fixup (tn-value symbol) :symbol-tls-index) thread-tn)
+                      value val-temp)))
+    (cond ((and (sc-is symbol immediate) (info :variable :wired-tls (tn-value symbol)))
+           ;; The TLS index is arbitrary but known to be nonzero,
+           ;; so we can resolve the displacement of the thread-local value
+           ;; at load-time, saving one instruction over the general case.
+           (inst lea cell (ea (make-fixup (tn-value symbol) :symbol-tls-index) thread-tn)))
+          (t
+           ;; These MOVs look the same, but when the symbol is immediate, this is
+           ;; a load from an absolute address. Needless to say, the names of these
+           ;; accessor macros are arbitrary - the difference is not very apparent.
+           (inst mov :dword cell (if (sc-is symbol immediate)
+                                     (symbol-tls-index-ea symbol)
+                                     (tls-index-of symbol)))
+           (inst add cell thread-tn)))
+    (inst cmp :qword (ea cell) no-tls-value-marker)
+    (inst jmp :ne STORE)
+    (emit-symbol-write-barrier symbol nil val-temp (vop-nth-arg 1 vop) value)
+    (get-symbol-value-slot-ea cell symbol)
+    STORE
+    (emit-store (ea cell) value val-temp)))
 
 (define-vop (fast-symbol-global-value)
   (:args (object :scs (descriptor-reg immediate)))
@@ -752,6 +759,29 @@
   funcallable-instance-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %funcallable-instance-info)
 
+;;; This is arguably a kludge without which you would hit the 'lose("Feh.")' in gencgc
+;;; if trying to fold funinstance setters into closure-index-set.
+;;; (Is's incorrect to manually manipulate the mark on a non-code page.)
+#-soft-card-marks
+(progn
+#-sb-xc-host
+(defun (setf %funcallable-instance-info) (newval fin index)
+  (%primitive %set-funinstance-info fin index newval)
+  newval)
+(define-vop (%set-funinstance-info)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg))
+         (value :scs (any-reg descriptor-reg)))
+  (:arg-types * tagged-num *)
+  (:temporary (:sc unsigned-reg) val-temp)
+  (:generator 4
+   (let ((ea (ea (- (* funcallable-instance-info-offset n-word-bytes) fun-pointer-lowtag)
+                 object index (index-scale n-word-bytes index))))
+     (pseudo-atomic ()
+       (emit-code-page-gengc-barrier object val-temp)
+       (emit-store ea value val-temp))))))
+
 (define-vop (closure-ref)
   (:args (object :scs (descriptor-reg)))
   (:results (value :scs (descriptor-reg any-reg)))
