@@ -143,8 +143,7 @@
   (:generator 6
     (move temp data)
     (inst shl temp (- n-widetag-bits n-fixnum-tag-bits))
-    ;; merge in the widetag. We should really preserve bit 63 as well
-    ;; which could be a GC mark bit, but it's not concurrent at the moment.
+    ;; merge in the widetag
     (inst mov :byte temp (ea (- other-pointer-lowtag) x))
     (storew temp x 0 other-pointer-lowtag)))
 (define-vop (logior-header-bits)
@@ -282,9 +281,7 @@
   (:results (result :scs (descriptor-reg)))
   (:generator 3
     (loadw result function closure-fun-slot fun-pointer-lowtag)
-    (inst lea result
-          (ea  (- fun-pointer-lowtag (* simple-fun-insts-offset n-word-bytes))
-               result))))
+    (inst sub result (- (* simple-fun-insts-offset n-word-bytes) fun-pointer-lowtag))))
 
 ;;;; symbol frobbing
 (defun load-symbol-dbinfo (result symbol)
@@ -519,7 +516,7 @@ number of CPU cycles elapsed as secondary value. EXPERIMENTAL."
    (inst store-coverage-mark index)))
 
 (define-vop ()
-  (:translate sb-lockless::get-next)
+  (:translate sb-lockless:get-next)
   (:policy :fast-safe)
   (:args (node :scs (descriptor-reg)))
   (:results (next-tagged :scs (descriptor-reg))
