@@ -43,59 +43,64 @@
 
 ;; Numerics including fixnum, excluding short-float. (INTEGER,RATIONAL)
 (defun %test-fixnum-and-headers (value temp target not-p headers
-                                 &key value-tn-ref)
+                                 &key value-tn-ref immediate-tested)
   (let ((drop-through (gen-label)))
     (case n-fixnum-tag-bits
-     (1 (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+     (1
+      (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+      (when (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                      (specifier-type 'fixnum))
         (inst test :byte temp 1)
-        (inst jmp :nz (if not-p drop-through target)) ; inverted
-        (%test-headers value temp target not-p nil headers
-                       :drop-through drop-through :compute-temp nil
-                       :value-tn-ref value-tn-ref))
+        (inst jmp :nz (if not-p drop-through target))) ; inverted
+      (%test-headers value temp target not-p nil headers
+                     :drop-through drop-through :compute-temp nil
+                     :value-tn-ref value-tn-ref
+                     :immediate-tested immediate-tested))
      (t
       (generate-fixnum-test value)
       (inst jmp :z (if not-p drop-through target))
       (%test-headers value temp target not-p nil headers
-                     :drop-through drop-through)))))
-
-;; I can see no reason this would ever be used.
-;; (or fixnum character|unbound-marker) is implausible.
-(defun %test-fixnum-and-immediate (value target not-p immediate)
-  (error "WAT")
-  (let ((drop-through (gen-label)))
-    (generate-fixnum-test value)
-    (inst jmp :z (if not-p drop-through target))
-    (%test-immediate value target not-p immediate drop-through)))
+                     :drop-through drop-through
+                     :immediate-tested immediate-tested)))))
 
 ;; Numerics
 (defun %test-fixnum-immediate-and-headers (value temp target not-p immediate headers
-                                           &key value-tn-ref)
+                                           &key value-tn-ref immediate-tested)
   (let ((drop-through (gen-label)))
     (case n-fixnum-tag-bits
-     (1 (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+     (1
+      (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+      (when (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                      (specifier-type 'fixnum))
         (inst test :byte temp 1)
-        (inst jmp :nz (if not-p drop-through target)) ; inverted
+        (inst jmp :nz (if not-p drop-through target))) ; inverted
+      (when (or (/= immediate single-float-widetag)
+                (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                          (specifier-type 'single-float)))
         (inst cmp :byte temp (- immediate other-pointer-lowtag))
-        (inst jmp :e (if not-p drop-through target))
-        (%test-headers value temp target not-p nil headers
-                       :drop-through drop-through :compute-temp nil
-                       :value-tn-ref value-tn-ref))
+        (inst jmp :e (if not-p drop-through target)))
+
+      (%test-headers value temp target not-p nil headers
+                     :drop-through drop-through :compute-temp nil
+                     :value-tn-ref value-tn-ref
+                     :immediate-tested immediate-tested))
      (t (generate-fixnum-test value)
         (inst jmp :z (if not-p drop-through target))
         (%test-immediate-and-headers value temp target not-p immediate headers
-                                     :drop-through drop-through)))))
+                                     :drop-through drop-through
+                                     :immediate-tested immediate-tested)))))
 
 (defun %test-immediate (value temp target not-p immediate
-                        &optional (drop-through (gen-label)))
-  (declare (ignore temp))
+                        &key value-tn-ref)
+  (declare (ignore temp value-tn-ref))
   (inst cmp :byte value immediate)
-  (inst jmp (if not-p :ne :e) target)
-  (emit-label drop-through))
+  (inst jmp (if not-p :ne :e) target))
 
 ;; Numerics including short-float, excluding fixnum
 (defun %test-immediate-and-headers (value temp target not-p immediate headers
                                     &key (drop-through (gen-label))
-                                         value-tn-ref)
+                                         value-tn-ref
+                                         immediate-tested)
   ;; Code a single instruction byte test if possible.
   (cond ((sc-is value any-reg descriptor-reg)
          (inst cmp :byte value immediate))
@@ -105,9 +110,11 @@
   (inst jmp :e (if not-p drop-through target))
   (%test-headers value temp target not-p nil headers
                  :drop-through drop-through
-                 :value-tn-ref value-tn-ref))
+                 :value-tn-ref value-tn-ref
+                 :immediate-tested immediate-tested))
 
-(defun %test-lowtag (value temp target not-p lowtag)
+(defun %test-lowtag (value temp target not-p lowtag &key value-tn-ref)
+  (declare (ignore value-tn-ref))
   (%lea-for-lowtag-test temp value lowtag)
   (inst test :byte temp lowtag-mask)
   (inst jmp (if not-p :nz :z) target))
@@ -117,7 +124,8 @@
                            (drop-through (gen-label))
                            (load-widetag t)
                            (compute-temp load-widetag)
-                           value-tn-ref)
+                           value-tn-ref
+                           immediate-tested)
   (let* ((lowtag (if function-p fun-pointer-lowtag other-pointer-lowtag))
          ;; It is preferable (smaller and faster code) to directly
          ;; compare the value in memory instead of loading it into
@@ -149,7 +157,7 @@
       (cond ((not load-widetag))
             ((and value-tn-ref
                   (eq lowtag other-pointer-lowtag)
-                  (other-pointer-tn-ref-p value-tn-ref t))) ; best case: lowtag is right
+                  (other-pointer-tn-ref-p value-tn-ref t immediate-tested))) ; best case: lowtag is right
             ((and value-tn-ref
                   ;; If HEADERS contains a range, then list pointers have to be
                   ;; disallowed - consider a list whose CAR has a fixnum that
@@ -366,7 +374,11 @@
 ;; the header of which must not overlap the static alloc regions
 #-sb-thread
 (aver (>= (- non-negative-fixnum-mask-constant-wired-address (* 2 n-word-bytes))
-          (+ (max boxed-region cons-region mixed-region) (* 3 n-word-bytes))))
+          (+ static-space-start
+             (max boxed-region-offset
+                  cons-region-offset
+                  mixed-region-offset)
+             (* 3 n-word-bytes))))
 
 ;;; An (unsigned-byte 64) can be represented with either a positive
 ;;; fixnum, a bignum with exactly one positive digit, or a bignum with
@@ -374,19 +386,26 @@
 (define-vop (unsigned-byte-64-p type-predicate)
   (:translate unsigned-byte-64-p)
   (:generator 10
-    (let ((not-target (gen-label))
-          (single-word (gen-label))
-          (fixnum-p (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum))))
+    (let* ((not-target (gen-label))
+           (single-word (gen-label))
+           (fixnum-p (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum)))
+           (not-signed-byte-64-p (not (types-equal-or-intersect (tn-ref-type args) (specifier-type 'signed-word))))
+           (unsigned-p (or not-signed-byte-64-p
+                           (not (types-equal-or-intersect (tn-ref-type args) (specifier-type '(integer * -1)))))))
       (multiple-value-bind (yep nope)
           (if not-p
               (values not-target target)
               (values target not-target))
         (when fixnum-p
-          ;; Is it a fixnum with the sign bit clear?
-          (inst test (ea non-negative-fixnum-mask-constant-wired-address) value)
-          (inst jmp :z yep))
+          (cond (unsigned-p
+                 (inst test :byte value fixnum-tag-mask)
+                 (inst jmp :z yep))
+                (t ;; Is it a fixnum with the sign bit clear?
+                 (inst test (ea non-negative-fixnum-mask-constant-wired-address) value)
+                 (inst jmp :z yep))))
         (cond ((fixnum-or-other-pointer-tn-ref-p args t)
-               (when fixnum-p
+               (when (and fixnum-p
+                          (not unsigned-p))
                  (inst test :byte value fixnum-tag-mask)
                  (inst jmp :z nope)))
               (t
@@ -395,9 +414,12 @@
                (inst jmp :ne nope)))
         ;; Get the header.
         (loadw temp value 0 other-pointer-lowtag)
-        ;; Is it one?
-        (inst cmp temp (bignum-header-for-length 1))
-        (inst jmp :e single-word)
+        (unless not-signed-byte-64-p
+          ;; Is it one?
+          (inst cmp temp (bignum-header-for-length 1))
+          (inst jmp :e (if unsigned-p
+                           yep
+                           single-word)))
         ;; If it's other than two, we can't be an (unsigned-byte 64)
         ;: Leave TEMP holding 0 in the affirmative case.
         (inst sub temp (bignum-header-for-length 2))
@@ -405,15 +427,19 @@
         ;; Compare the second digit to zero (in TEMP).
         (inst cmp (object-slot-ea value (1+ bignum-digits-offset) other-pointer-lowtag)
               temp)
-        (inst jmp :z yep) ; All zeros, its an (unsigned-byte 64).
-        (inst jmp nope)
+        (cond (unsigned-p
+               (inst jmp (if not-p :nz :z) target))
+              (t
+               (inst jmp :z yep) ; All zeros, its an (unsigned-byte 64).
+               (inst jmp nope)))
 
-        (emit-label single-word)
-        ;; Get the single digit.
-        (loadw temp value bignum-digits-offset other-pointer-lowtag)
-        ;; positive implies (unsigned-byte 64).
-        (inst test temp temp)
-        (inst jmp (if not-p :s :ns) target)
+        (unless unsigned-p
+          (emit-label single-word)
+          ;; Get the single digit.
+          (loadw temp value bignum-digits-offset other-pointer-lowtag)
+          ;; positive implies (unsigned-byte 64).
+          (inst test temp temp)
+          (inst jmp (if not-p :s :ns) target))
 
         (emit-label not-target)))))
 
@@ -932,15 +958,14 @@
 (define-vop (load-other-pointer-widetag)
   (:args (value :scs (any-reg descriptor-reg)))
   (:arg-refs args)
-  (:info not-other-pointer-label null-label)
+  (:info not-other-pointer-label null-label zero-extend)
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
   (:generator 1
-    ;; FIXME: can we pass T as the 2nd arg to OTHER-POINTER-TN-REF-P here?
-    ;; I'm confused as to the intent of NULL-LABEL.
-    ;; Must branching occur if nil, or _may_ branching occur?
-    (cond ((other-pointer-tn-ref-p args)
-           (inst mov :byte r (ea (- other-pointer-lowtag) value)))
+    (cond ((other-pointer-tn-ref-p args (not null-label))
+           (if zero-extend
+               (inst movzx '(:byte :dword) r (ea (- other-pointer-lowtag) value))
+               (inst mov :byte r (ea (- other-pointer-lowtag) value))))
           (t
            (when null-label
              (inst cmp value nil-value)
@@ -948,7 +973,9 @@
            (%lea-for-lowtag-test r value other-pointer-lowtag :qword)
            (inst test :byte r lowtag-mask)
            (inst jmp :nz not-other-pointer-label)
-           (inst mov :byte r (ea r))))))
+           (if zero-extend
+               (inst movzx '(:byte :dword) r (ea r))
+               (inst mov :byte r (ea r)))))))
 
 (define-vop (test-widetag)
   (:args (value :scs (unsigned-reg) :target temp))
@@ -1083,7 +1110,6 @@
   (:args (layout :scs (descriptor-reg)))
   (:arg-types * (:constant t))
   (:policy :fast-safe)
-  (:conditional)
   (:info target not-p test-layout)
   (:generator 4
     (structure-is-a layout test-layout target not-p done)
