@@ -9,7 +9,8 @@
  * files for more information.
  */
 
-#include "sbcl.h"
+#include "genesis/sbcl.h"
+#include "lispobj.h"
 
 #include <stdio.h>
 #include <sys/types.h>
@@ -34,11 +35,14 @@
 #include "lispregs.h"
 #include "interrupt.h"
 #include "thread.h"
+#include "genesis/instance.h"
 #include "genesis/static-symbols.h"
 #include "genesis/primitive-objects.h"
 #include "genesis/gc-tables.h"
-#include "gc-internal.h"
-
+#include "genesis/unwind-block.h"
+#include "gc.h"
+#include "../../tlsf-bsd/tlsf/tlsf.h"
+extern void* tlsf_control;
 
 /* When we need to do command input, we use this stream, which is not
  * in general stdin, so that things will "work" (as well as being
@@ -59,20 +63,26 @@ struct crash_preamble {
     uword_t signature;
     uword_t static_start;
     uword_t static_nbytes;
+    uword_t readonly_start;
+    uword_t readonly_nbytes;
     uword_t dynspace_start;
     long dynspace_npages;
     int card_size;
     int card_table_nbits;
     // fixedobj data dumped: pages, page table
     uword_t fixedobj_start, fixedobj_size, fixedobj_free_pointer;
-    // varyobj data dumped: pages, touched_bits, page table
-    uword_t varyobj_start, varyobj_size, varyobj_free_pointer;
+    // text data dumped: pages, touched_bits, page table
+    uword_t text_start, text_size;
+    lispobj *tlsf_mem_start, *text_space_highwatermark;
+    lispobj sentinel_block[3];
+    void* tlsf_control_address;
     int nthreads;
     int tls_size;
     lispobj lisp_package_vector;
+    int sizeof_context;
+    int tlsf_control_size;
     char sprof_enabled;
     char pin_dynspace_code;
-    int sizeof_context;
 };
 struct crash_thread_preamble {
     uword_t address;
@@ -82,20 +92,21 @@ struct crash_thread_preamble {
 };
 
 // Prevent some mixups in case you add fields to the crash dump
-const int CRASH_PREAMBLE_SIGNATURE =
+const uword_t CRASH_PREAMBLE_SIGNATURE =
     (sizeof (struct crash_preamble) << 16) | sizeof (struct crash_thread_preamble);
 
-#if defined LISP_FEATURE_X86_64 // un-tested elsewhere
+#if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64 // un-tested elsewhere
 #include <errno.h>
-static void checked_write(int fd, void* buf, long nbytes)
+static void checked_write(char *legend, int fd, void* buf, long nbytes)
 {
     ssize_t wrote = write(fd, buf, nbytes);
     if (wrote != nbytes) lose("short write, errno=%d", errno);
+    fprintf(stderr, "%s: %lx bytes\n", legend, nbytes);
 }
 
 #include "immobile-space.h"
 void save_gc_crashdump(char *pathname,
-                       uword_t approx_stackptr_at_gc_start)
+                       lispobj* cur_thread_approx_stackptr)
 {
     extern int pin_all_dynamic_space_code;
     int fd = open(pathname, O_WRONLY|O_CREAT, 0666);
@@ -105,14 +116,12 @@ void save_gc_crashdump(char *pathname,
     fprintf(stderr, "save: %d threads\n", nthreads);
     struct crash_preamble preamble;
     unsigned long nbytes_heap = next_free_page * GENCGC_PAGE_BYTES;
-#ifdef LISP_FEATURE_SB_THREAD
     int nbytes_tls = SymbolValue(FREE_TLS_INDEX,0);
-#else
-    int nbytes_tls = sizeof (struct thread);
-#endif
     preamble.signature = CRASH_PREAMBLE_SIGNATURE;
     preamble.static_start = STATIC_SPACE_START;
     preamble.static_nbytes = (uword_t)static_space_free_pointer - STATIC_SPACE_START;
+    preamble.readonly_start = READ_ONLY_SPACE_START;
+    preamble.readonly_nbytes = (uword_t)read_only_space_free_pointer - READ_ONLY_SPACE_START;
     preamble.dynspace_start = DYNAMIC_SPACE_START;
     preamble.dynspace_npages = next_free_page;
     preamble.card_size = GENCGC_CARD_BYTES;
@@ -124,33 +133,43 @@ void save_gc_crashdump(char *pathname,
     preamble.pin_dynspace_code = pin_all_dynamic_space_code;
     preamble.sizeof_context = sizeof (os_context_t);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
+    char *tlsf_memory_end = (char*)TEXT_SPACE_START + text_space_size;
     preamble.fixedobj_start = FIXEDOBJ_SPACE_START;
     preamble.fixedobj_size = FIXEDOBJ_SPACE_SIZE;
     preamble.fixedobj_free_pointer = (uword_t)fixedobj_free_pointer;
-    preamble.varyobj_start = VARYOBJ_SPACE_START;
-    preamble.varyobj_size = VARYOBJ_SPACE_SIZE;
-    preamble.varyobj_free_pointer = (uword_t)varyobj_free_pointer;
+    preamble.text_start = TEXT_SPACE_START;
+    preamble.text_size = text_space_size;
+    preamble.text_space_highwatermark = text_space_highwatermark;
+    preamble.tlsf_mem_start = tlsf_mem_start;
+    preamble.tlsf_control_address = tlsf_control;
+    preamble.tlsf_control_size = tlsf_size();
+    memcpy(preamble.sentinel_block, tlsf_memory_end-3*N_WORD_BYTES, 3*N_WORD_BYTES);
 #endif
-    // write the preamble and static space
-    checked_write(fd, &preamble, sizeof preamble);
-    checked_write(fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
-    // write the dynamic-space, PTEs, card table
+    // write the preamble and static + readonly spaces
+    checked_write("preamble", fd, &preamble, sizeof preamble);
+    checked_write("static", fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
+    checked_write("R/O", fd, (char*)READ_ONLY_SPACE_START, preamble.readonly_nbytes);
 
-    fprintf(stderr, "Writing dynamic space\n");
-    checked_write(fd, (char*)DYNAMIC_SPACE_START, nbytes_heap);
-    checked_write(fd, page_table, sizeof (struct page) * next_free_page);
-    checked_write(fd, gc_card_mark, 1+gc_card_table_mask);
+    // write the dynamic-space, PTEs, card table
+    checked_write("dynamic", fd, (char*)DYNAMIC_SPACE_START, nbytes_heap);
+    checked_write("PTE", fd, page_table, sizeof (struct page) * next_free_page);
+    checked_write("cardmark", fd, gc_card_mark, 1+gc_card_table_mask);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     int usage = (uword_t)fixedobj_free_pointer - FIXEDOBJ_SPACE_START;
-    checked_write(fd, (char*)FIXEDOBJ_SPACE_START, usage);
+    checked_write("fixedobj", fd, (char*)FIXEDOBJ_SPACE_START, usage);
     int total_npages = FIXEDOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
-    checked_write(fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
-    usage = (uword_t)varyobj_free_pointer - VARYOBJ_SPACE_START;
-    checked_write(fd, (char*)VARYOBJ_SPACE_START, usage);
-    total_npages = VARYOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
+    checked_write("fixedobj_PTE", fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
+    usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
+    // write the block_header_t that is just beyond the high water mark
+    checked_write("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
+    total_npages = text_space_size / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
-    checked_write(fd, varyobj_page_touched_bits, n_bitmap_elts * sizeof (int));
-    checked_write(fd, varyobj_pages, total_npages * sizeof (int));
+    checked_write("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
+    checked_write("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
+    checked_write("TLSF_control", fd, tlsf_control, preamble.tlsf_control_size);
+    int tlsf_memory_size = tlsf_memory_end - (char*)tlsf_mem_start;
+    int n_tlsf_pages = tlsf_memory_size / IMMOBILE_CARD_BYTES;
+    checked_write("TLSF_sso", fd, tlsf_page_sso, n_tlsf_pages * sizeof (short));
 #endif
     struct crash_thread_preamble thread_preamble;
     for_each_thread(th) {
@@ -160,7 +179,7 @@ void save_gc_crashdump(char *pathname,
         if (ici) {
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
             sp = *os_context_register_addr(threadcontext, reg_SP);
-#else
+#elif defined reg_CSP
             sp = *os_context_register_addr(threadcontext, reg_CSP);
 #endif
         } else if (th->state_word.state == STATE_DEAD) {
@@ -177,7 +196,7 @@ void save_gc_crashdump(char *pathname,
               _exit(1);
             }
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
-            sp = approx_stackptr_at_gc_start;
+            sp = (uword_t)cur_thread_approx_stackptr;
 #else
             sp = access_control_stack_pointer(th);
 #endif
@@ -197,18 +216,18 @@ void save_gc_crashdump(char *pathname,
         thread_preamble.control_stack_nbytes = nbytes_control_stack;
         thread_preamble.binding_stack_nbytes = nbytes_binding_stack;
         // write the preamble
-        checked_write(fd, &thread_preamble, sizeof thread_preamble);
+        checked_write("thread", fd, &thread_preamble, sizeof thread_preamble);
         // write 0 or 1 contexts, control-stack, binding-stack, TLS
-        if (ici) write(fd, threadcontext, preamble.sizeof_context);
+        if (ici) checked_write(" ctxt", fd, threadcontext, preamble.sizeof_context);
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
-        checked_write(fd, (char*)sp, nbytes_control_stack);
+        checked_write(" stack", fd, (char*)sp, nbytes_control_stack);
 #else
-        checked_write(fd, th->control_stack_start, nbytes_control_stack);
+        checked_write(" stack", fd, th->control_stack_start, nbytes_control_stack);
 #endif
-        checked_write(fd, th->binding_stack_start, nbytes_binding_stack);
-        checked_write(fd, th, nbytes_tls);
+        checked_write(" bindings", fd, th->binding_stack_start, nbytes_binding_stack);
+        checked_write(" TLS", fd, th, nbytes_tls);
     }
-    checked_write(fd, "SB.Crash", 8); // trailing signature
+    checked_write("sig", fd, "SB.Crash", 8); // trailing signature
     close(fd);
 }
 #endif
@@ -217,62 +236,136 @@ static cmd call_cmd, dump_cmd, print_cmd, quit_cmd, help_cmd;
 static cmd flush_cmd, regs_cmd, exit_cmd;
 static cmd print_context_cmd, pte_cmd, search_cmd;
 static cmd backtrace_cmd, catchers_cmd;
-static cmd threads_cmd;
+static cmd threads_cmd, findpath_cmd, layouts_cmd;
 
 extern void gc_stop_the_world(), gc_start_the_world();
 static void suspend_other_threads() {
-#ifdef LISP_FEATURE_SB_THREAD
     gc_stop_the_world();
-#endif
     // It might make sense for each thread's stop-for-gc handler to close its region
     // versus doing this loop
     struct thread *th;
-    for_each_thread(th) { gc_close_thread_regions(th); }
+    for_each_thread(th) { gc_close_thread_regions(th, 0); }
     gc_close_collector_regions(0);
 }
 static void unsuspend_other_threads() {
-#ifdef LISP_FEATURE_SB_THREAD
     gc_start_the_world();
-#endif
 }
 
 static int save_cmd(char **ptr) {
-#if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_SB_THREAD
     char *name  = parse_token(ptr);
     if (!name) {
         fprintf(stderr, "Need filename\n");
-        return 1;
+        return 0;
     }
+#if (defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64) && defined LISP_FEATURE_SB_THREAD
     suspend_other_threads();
-    save_gc_crashdump(name, (uword_t)__builtin_frame_address(0));
+    save_gc_crashdump(name, (lispobj*)__builtin_frame_address(0));
     unsuspend_other_threads();
 #else
     fprintf(stderr, "Unimplemented\n");
 #endif
     return 0;
 }
-
-static int threads_cmd(char __attribute__((unused)) **ptr) {
+void list_lisp_threads(int regions) {
     struct thread* th;
-    fprintf(stderr, "(thread*,pthread,sb-vm:thread)\n");
+    fprintf(stderr, "(thread*,pthread,sb-vm:thread,name)\n");
     void* pthread;
     for_each_thread(th) {
         memcpy(&pthread, &th->os_thread, N_WORD_BYTES);
-        fprintf(stderr, "%p %p %p\n", th, pthread, (void*)th->lisp_thread);
+        struct thread_instance* i = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
+        lispobj name = instancep(th->lisp_thread) ? i->_name : 0;
+        char* cname = NULL;
+        // it's OK to call widetag_of on NIL but not NULL
+        if (name && simple_base_string_p(name)) cname = vector_sap(name);
+        fprintf(stderr, "%p %p %p \"%s\"\n", th, pthread, (void*)i, cname);
+        if (regions) {
+#define show_tlab(label, r) fprintf(stderr, "  %s @ %p: %p %p %p\n", label, \
+         &th->r, th->r.start_addr, th->r.end_addr, th->r.free_pointer)
+            show_tlab("usr cons ", cons_tlab);
+            show_tlab("usr mix  ", mixed_tlab);
+            show_tlab("sys cons ", sys_cons_tlab);
+            show_tlab("sys mix  ", sys_mixed_tlab);
+#undef show_tlab
+        }
     }
+    if (regions) {
+#define show_tlab(label, r) fprintf(stderr, "  %s @ %p: %p %p %p\n", label, \
+         r, r->start_addr, r->end_addr, r->free_pointer)
+        fprintf(stderr, "global regions:\n");
+        show_tlab("mixed    ", mixed_region);
+        show_tlab("smallmix ", small_mixed_region);
+        show_tlab("unboxed  ", unboxed_region);
+        show_tlab("code     ", code_region);
+        show_tlab("boxed    ", boxed_region);
+        show_tlab("cons     ", cons_region);
+#undef show_tlab
+    }
+}
+static int threads_cmd(char **ptr) {
+    list_lisp_threads(more_p(ptr) && !strncmp(*ptr, "-r", 2));
+    return 0;
+}
+extern int heap_trace_verbose;
+extern int gc_pathfind_aux(lispobj*, lispobj, lispobj, lispobj, int);
+
+static int findpath_cmd(char **ptr) {
+    // prevent the path finder from seeing the object that results from parsing the command
+    lispobj* stackptr = (lispobj*)&ptr;
+    // overaligned for 32-bit but doesn't matter
+    struct vector __attribute__ ((aligned (16))) result;
+    struct cons __attribute__ ((aligned (16))) list;
+    /* Despite capturing the stack pointer coming in to this function,
+     * it's very difficult to ensure that a locally allocated weak pointer
+     * is not above that (and hence its value slot seen).
+     * So it has to be malloc()ed. */
+    char* wp_mem = malloc(sizeof (struct weak_pointer) + N_WORD_BYTES);
+    struct weak_pointer* wp =
+      (void*)(wp_mem + (((uword_t)wp_mem & LOWTAG_MASK) ? N_WORD_BYTES : 0));
+    wp->header = ((WEAK_POINTER_SIZE-1)<<N_WIDETAG_BITS)|WEAK_POINTER_WIDETAG;
+    if (parse_lispobj(ptr, &wp->value)) {
+        list.car = make_lispobj(wp, OTHER_POINTER_LOWTAG);
+        list.cdr = NIL;
+        result.header = SIMPLE_VECTOR_WIDETAG;
+        result.length_ = make_fixnum(1);
+        result.data[0] = 0;
+        suspend_other_threads();
+        int save = heap_trace_verbose;
+        heap_trace_verbose = 4;
+        gc_pathfind_aux(stackptr,
+                        make_lispobj(&list, LIST_POINTER_LOWTAG),
+                        make_lispobj(&result, OTHER_POINTER_LOWTAG),
+                        0, 2);
+        heap_trace_verbose = save;
+        unsuspend_other_threads();
+        lispobj path = result.data[0];
+        if (listp(path)) {
+            fprintf(stderr, "Answer:\n");
+            while (path != NIL) {
+                struct cons* pair = CONS(CONS(path)->car);
+                if (listp(pair->cdr)) {
+                    // thread root - complicated to print
+                } else {
+                    // otherwise, object and word index
+                    fprintf(stderr, " %"OBJ_FMTX" word %d\n",
+                            pair->car, (int)pair->cdr);
+                }
+                path = CONS(path)->cdr;
+            }
+        }
+    }
+    free(wp_mem);
     return 0;
 }
 static int verify_cmd(char __attribute__((unused)) **ptr) {
     gencgc_verbose = 1;
     suspend_other_threads();
-    verify_heap(0);
+    verify_heap(0, 0);
     unsuspend_other_threads();
     return 0;
 }
 static int gc_cmd(char **ptr) {
     int last_gen = 0;
     extern generation_index_t verify_gens;
-    extern boolean pre_verify_gen_0;
     if (more_p(ptr)) parse_number(ptr, &last_gen);
     gencgc_verbose = 2;
     pre_verify_gen_0 = 1;
@@ -283,10 +376,22 @@ static int gc_cmd(char **ptr) {
     return 0;
 }
 
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+static int tlsf_cmd(__attribute__((unused)) char **ptr) {
+    tlsf_dump_pool(tlsf_control, tlsf_mem_start, "/dev/tty");
+#ifdef TLSF_CONFIG_DEBUG
+    tlsf_check(tlsf_control);
+    tlsf_check_pool(tlsf_mem_start);
+#endif
+    return 0;
+}
+#endif
+
 static struct cmd {
     char *cmd, *help;
     int (*fn)(char **ptr);
 } supported_cmds[] = {
+    // Commands with no help string are all at-your-own-risk
     {"help", "Display this help information.", help_cmd},
     {"?", "(an alias for help)", help_cmd},
     {"backtrace", "Backtrace up to N frames.", backtrace_cmd},
@@ -296,17 +401,22 @@ static struct cmd {
     {"dump", "Dump memory starting at ADDRESS for COUNT words.", dump_cmd},
     {"d", "(an alias for dump)", dump_cmd},
     {"exit", "Exit this instance of the monitor.", exit_cmd},
+    {"findpath", "Find path to an object.", findpath_cmd},
     {"flush", "Flush all temp variables.", flush_cmd},
+    {"layouts", "Dump LAYOUT instances.", layouts_cmd},
     {"print", "Print object at ADDRESS.", print_cmd},
     {"p", "(an alias for print)", print_cmd},
     {"pte", "Page table entry for address", pte_cmd},
     {"quit", "Quit.", quit_cmd},
     {"regs", "Display current Lisp registers.", regs_cmd},
     {"search", "Search heap for object.", search_cmd},
-    {"save", "Produce crashdump", save_cmd},
+    {"save", 0, save_cmd}, // snapshot heap to file ("best effort" though)
     {"threads", "List threads", threads_cmd},
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    {"tlsfdump", 0, tlsf_cmd}, // (unsafely) dump TLSF structures
+#endif
     {"verify", "Check heap invariants", verify_cmd},
-    {"gc", "Collect garbage", gc_cmd},
+    {"gc", 0, gc_cmd}, // (this is not the right way to invoke GC)
     {NULL, NULL, NULL}
 };
 
@@ -319,7 +429,7 @@ visible(unsigned char c)
         return c;
 }
 
-static boolean valid_widetag_p(unsigned char widetag) {
+static bool valid_widetag_p(unsigned char widetag) {
     // TODO: ensure that widetag is defined (not "unused") and is for a headered object
     // (i.e. is not CHARACTER_WIDETAG and not some other things)
     return other_immediate_lowtag_p(widetag);
@@ -363,7 +473,7 @@ dump_cmd(char **ptr)
         count = -count;
     }
 
-    boolean aligned = ((uword_t)addr & LOWTAG_MASK) == 0;
+    bool aligned = ((uword_t)addr & LOWTAG_MASK) == 0;
     if (decode && (!aligned || displacement < 0)) {
         printf("Sorry, can only decode if aligned and stepping forward\n");
         decode = 0;
@@ -394,13 +504,22 @@ dump_cmd(char **ptr)
                    visible(cptr[0]), visible(cptr[1]),
                    visible(cptr[2]), visible(cptr[3]));
 #endif
-#ifdef LISP_FEATURE_GENCGC
+#ifdef LISP_FEATURE_GENERATIONAL
             if (aligned) {
                 lispobj ptr = *(lispobj*)addr;
                 int gen;
                 if (is_lisp_pointer(ptr) && gc_managed_heap_space_p(ptr)
-                    && (gen = gc_gen_of(ptr, 99)) != 99) // say that static is 99
+                    && (gen = gc_gen_of(ptr, 99)) != 99) { // say that static is 99
                     if (gen != 99) printf(" | %d", gen);
+                } else {
+                    printf("    "); // padding to make MR part line up
+                }
+            }
+#endif
+#ifdef LISP_FEATURE_MARK_REGION_GC
+            if (aligned && find_page_index(addr) != -1) {
+                extern bool allocation_bit_marked(void*);
+                printf(" %c", allocation_bit_marked(addr) ? '*' : ' ');
             }
 #endif
             if (decode && addr == (char*)next_object) {
@@ -449,7 +568,7 @@ pte_cmd(char **ptr)
 static int
 regs_cmd(char __attribute__((unused)) **ptr)
 {
-    struct thread __attribute__((unused)) *thread=get_sb_vm_thread();
+    struct thread __attribute__((unused)) *thread = get_sb_vm_thread();
 
     printf("CSP\t=\t%p   ", access_control_stack_pointer(thread));
 #if !defined(LISP_FEATURE_X86) && !defined(LISP_FEATURE_X86_64)
@@ -463,17 +582,13 @@ regs_cmd(char __attribute__((unused)) **ptr)
     printf("\n");
 #endif
 
-#ifdef LISP_FEATURE_GENCGC
+#ifdef LISP_FEATURE_GENERATIONAL
     /* printf("DYNAMIC\t=\t%p\n", (void*)DYNAMIC_SPACE_START); */
 #else
     printf("STATIC\t=\t%p   ", static_space_free_pointer);
     printf("RDONLY\t=\t%p   ", read_only_space_free_pointer);
     printf("DYNAMIC\t=\t%p\n", (void*)current_dynamic_space);
 #endif
-
-#ifndef LISP_FEATURE_GENCGC
-    printf("TRIGGER\t=\t%p\n", (void*)current_auto_gc_trigger);
-#endif
     return 0;
 }
 
@@ -600,6 +715,10 @@ print_context(os_context_t *context)
         brief_print((lispobj)(*os_context_register_addr(context,i)));
 
     }
+#if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_PPC)
+    printf("DAR:\t\t 0x%08lx\n", (unsigned long)(*os_context_register_addr(context, 41)));
+    printf("DSISR:\t\t 0x%08lx\n", (unsigned long)(*os_context_register_addr(context, 42)));
+#endif
 #ifndef REG_PC
     printf("PC:\t\t  0x%08lx\n", (unsigned long)os_context_pc(context));
 #endif
@@ -609,7 +728,7 @@ static int
 print_context_cmd(char **ptr)
 {
     int free_ici;
-    struct thread *thread=get_sb_vm_thread();
+    struct thread *thread = get_sb_vm_thread();
 
     free_ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,thread));
 
@@ -693,7 +812,72 @@ catchers_cmd(char __attribute__((unused)) **ptr)
     return 0;
 }
 
-extern boolean gc_active_p;
+struct layout_collection {
+    struct cons* list;
+    int passno;
+};
+static int count_layout_occurs(lispobj x, struct cons* list)
+{
+    int ct = 0;
+    for ( ; list ; list = (void*)list->cdr ) if (list->car == x) ++ct;
+    return ct;
+}
+
+static uword_t display_layouts(lispobj* where, lispobj* limit, uword_t arg)
+{
+    extern struct vector * classoid_name(lispobj * classoid);
+    struct layout_collection *lc = (void*)arg;
+    where = next_object(where, 0, limit); /* find first marked object */
+    for ( ; where ; where = next_object(where, object_size(where), limit) ) {
+        if (widetag_of(where) == INSTANCE_WIDETAG &&
+            instance_layout(where) != 0 &&
+            layout_depth2_id(LAYOUT(instance_layout(where))) == LAYOUT_LAYOUT_ID) {
+            struct layout* l = (void*)where;
+            struct classoid* c = (void*)native_pointer(l->classoid);
+            if (lc->passno == 1) {
+                // on the first pass, just collect the classoids;
+                struct cons* cons = malloc(sizeof (struct cons));
+                cons->car = (lispobj)c;
+                cons->cdr = (lispobj)lc->list;
+                lc->list = cons;
+            } else {
+                // print some information preceded by a '*' if more than one
+                // layout points to the classoid of this layout
+                int count = count_layout_occurs((lispobj)c, lc->list);
+                struct vector* v = classoid_name((lispobj*)c);
+                char* name =
+                  header_widetag(v->header)==SIMPLE_BASE_STRING_WIDETAG ? (char*)v->data : "?";
+                fprintf(stderr, "%c %p %16" OBJ_FMTX " %16" OBJ_FMTX " %p %" OBJ_FMTX " %s\n",
+                        count>1 ? '*' : ' ', l, l->clos_hash, l->uw_id_word0,
+                        c, l->invalid, name);
+
+            }
+        }
+    }
+   return 0;
+}
+
+static int layouts_cmd(char __attribute__((unused)) **ptr)
+{
+    fprintf(stderr, "Dup, Layout, Hash, ID_Word, Classoid, Invalid, Name\n");
+    struct layout_collection lc;
+    lc.list = 0;
+    for (lc.passno = 1; lc.passno <= 2; ++lc.passno) {
+        walk_generation(display_layouts, -1, (uword_t)&lc);
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+        display_layouts((lispobj*)FIXEDOBJ_SPACE_START, fixedobj_free_pointer,
+                        (uword_t)&lc);
+#endif
+    }
+    struct cons* l = lc.list;
+    while (l) {
+        struct cons* next = (struct cons*)l->cdr;
+        free(l);
+        l = next;
+    }
+    return 0;
+}
+
 extern FILE *gc_activitylog_file;
 void
 ldb_monitor(void)
@@ -757,41 +941,41 @@ ldb_monitor(void)
     }
 }
 
-/* what we do when things go badly wrong at a low level */
-void
-monitor_or_something()
-{
-    ldb_monitor();
-}
-
 #ifdef STANDALONE_LDB
 void gc_stop_the_world() { } // do nothing
 void gc_start_the_world() { } // do nothing
 #include <errno.h>
 #include "core.h"
-#include "gencgc-private.h"
 struct lisp_startup_options lisp_startup_options;
+char gc_coalesce_string_literals;
 
 void unwind_binding_stack() { lose("Can't unwind binding stack"); }
-FILE *prepare_to_save(char *filename, boolean prepend_runtime, void **runtime_bytes,
-                      size_t *runtime_size) {
+FILE *prepare_to_save(__attribute__((unused)) char *filename,
+                      __attribute__((unused)) bool prepend_runtime,
+                      __attribute__((unused)) void **runtime_bytes,
+                      __attribute__((unused)) size_t *runtime_size) {
     lose("Can't prepare_to_save");
 }
-boolean save_runtime_to_filehandle(FILE *output, void *runtime, size_t runtime_size,
-                                   int application_type) {
+bool save_runtime_to_filehandle(__attribute__((unused)) FILE *output,
+                                __attribute__((unused)) void *runtime,
+                                __attribute__((unused)) size_t runtime_size,
+                                __attribute__((unused)) int application_type) {
     lose("Can't save_runtime_to_filehandle");
 }
-boolean save_to_filehandle(FILE *file, char *filename, lispobj init_function,
-                           boolean make_executable,
-                           boolean save_runtime_options,
-                           int core_compression_level) {
+bool save_to_filehandle(__attribute__((unused)) FILE *file,
+                        __attribute__((unused)) char *filename,
+                        __attribute__((unused)) lispobj init_function,
+                        __attribute__((unused)) bool make_executable,
+                        __attribute__((unused)) bool save_runtime_options,
+                        __attribute__((unused)) int core_compression_level) {
     lose("Can't save_to_filehandle");
 }
 
-static size_t checked_read(int fd, void* buf, size_t n)
+static size_t checked_read(char *legend, int fd, void* buf, size_t n)
 {
     size_t result = read(fd, buf, n);
     if (result != n) { lose("read failed, errno=%d", errno); }
+    if (legend) fprintf(stderr, "%s: %zu bytes\n", legend, n);
     return result;
 }
 
@@ -810,7 +994,7 @@ char *pagetypedesc(int type)
 extern void gc_allocate_ptes();
 extern void recompute_gen_bytes_allocated();
 extern void print_generation_stats();
-extern struct thread *alloc_thread_struct(void*,lispobj);
+extern struct thread *alloc_thread_struct(void*);
 
 int load_gc_crashdump(char* pathname)
 {
@@ -825,9 +1009,9 @@ int load_gc_crashdump(char* pathname)
     }
     struct crash_preamble preamble;
     struct crash_thread_preamble thread_preamble;
-    checked_read(fd, &preamble, sizeof preamble);
-    printf("static=%lx nbytes=%x\n", preamble.static_start, (int)preamble.static_nbytes);
-    printf("heap_start=%lx npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
+    checked_read("preamble", fd, &preamble, sizeof preamble);
+    printf("static=%"OBJ_FMTX" nbytes=%x\n", preamble.static_start, (int)preamble.static_nbytes);
+    printf("heap_start=%"OBJ_FMTX" npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
     // pin_dynspace_code is for display only. It gets recomputed as the
     // logical OR of all threads' values of *GC-PIN-CODE-PAGES*.
     printf("sprof_enabled=%d pin_dynspace_code=%d packages=%p\n",
@@ -836,93 +1020,119 @@ int load_gc_crashdump(char* pathname)
     lisp_package_vector = preamble.lisp_package_vector;
     sb_sprof_enabled = preamble.sprof_enabled;
     if (preamble.signature != CRASH_PREAMBLE_SIGNATURE)
-        lose("Can't load crashdump: bad header (have %lx, expect %lx)",
-             preamble.signature, (long)CRASH_PREAMBLE_SIGNATURE);
+        lose("Can't load crashdump: bad header (have %"OBJ_FMTX", expect %"OBJ_FMTX")",
+             preamble.signature, (uword_t)CRASH_PREAMBLE_SIGNATURE);
     if (preamble.card_size != GENCGC_CARD_BYTES)
         lose("Can't load crashdump: memory parameters differ");
     gc_card_table_nbits = preamble.card_table_nbits;
     gc_allocate_ptes();
     next_free_page = preamble.dynspace_npages;
-    checked_read(fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
+    // static + readonly
+    checked_read("static", fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
     static_space_free_pointer = (lispobj*)(STATIC_SPACE_START + preamble.static_nbytes);
+    if (preamble.readonly_nbytes) {
+      // READ_ONLY_SPACE_START = preamble.readonly_start;
+      os_alloc_gc_space(READ_ONLY_CORE_SPACE_ID, 0, (char*)preamble.readonly_start,
+                        ALIGN_UP(preamble.readonly_nbytes, 4096));
+      checked_read("R/O", fd, (char*)preamble.readonly_start, preamble.readonly_nbytes);
+    }
+    read_only_space_free_pointer = (lispobj*)(READ_ONLY_SPACE_START + preamble.readonly_nbytes);
+    //
     DYNAMIC_SPACE_START = preamble.dynspace_start;
     long dynspace_nbytes = preamble.dynspace_npages * GENCGC_PAGE_BYTES;
-    char* dynspace = os_validate(0, (char*)preamble.dynspace_start,
-                                 DEFAULT_DYNAMIC_SPACE_SIZE, 0, 0);
+    char* dynspace = os_alloc_gc_space(DYNAMIC_CORE_SPACE_ID, 0, (char*)preamble.dynspace_start,
+                                       DEFAULT_DYNAMIC_SPACE_SIZE);
     if (dynspace != (char*)preamble.dynspace_start)
         lose("Didn't map dynamic space where expected: %p vs %p",
              dynspace, (char*)preamble.dynspace_start);
-    checked_read(fd, (char*)DYNAMIC_SPACE_START, dynspace_nbytes);
-    fprintf(stderr, "snapshot: %ld pages in use (%ld bytes)\n",
+    checked_read("dynamic", fd, (char*)DYNAMIC_SPACE_START, dynspace_nbytes);
+    fprintf(stderr, "snapshot: %"PRIdPTR" pages in use (%ld bytes)\n",
             next_free_page, dynspace_nbytes);
-    checked_read(fd, page_table, sizeof (struct page) * next_free_page);
+    checked_read("PTE", fd, page_table, sizeof (struct page) * next_free_page);
     recompute_gen_bytes_allocated();
-    checked_read(fd, gc_card_mark, 1+gc_card_table_mask);
+    checked_read("cardmark", fd, gc_card_mark, 1+gc_card_table_mask);
     print_generation_stats();
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     extern void gc_init_immobile(),
                 calc_immobile_space_bounds(),
                 write_protect_immobile_space();
     gc_assert(preamble.fixedobj_size == FIXEDOBJ_SPACE_SIZE);
-    gc_assert(preamble.varyobj_size == VARYOBJ_SPACE_SIZE);
+    gc_assert(preamble.text_size == text_space_size);
     FIXEDOBJ_SPACE_START = preamble.fixedobj_start;
-    VARYOBJ_SPACE_START = preamble.varyobj_start;
+    TEXT_SPACE_START = preamble.text_start;
     fixedobj_free_pointer = (lispobj*)preamble.fixedobj_free_pointer;
-    varyobj_free_pointer = (lispobj*)preamble.varyobj_free_pointer;
-    os_validate(0, (char*)FIXEDOBJ_SPACE_START, FIXEDOBJ_SPACE_SIZE, 0, 0);
-    os_validate(0, (char*)VARYOBJ_SPACE_START, VARYOBJ_SPACE_SIZE, 0, 0);
+    text_space_highwatermark = (lispobj*)preamble.text_space_highwatermark;
+    os_alloc_gc_space(IMMOBILE_FIXEDOBJ_CORE_SPACE_ID, 0, (char*)FIXEDOBJ_SPACE_START,
+                      FIXEDOBJ_SPACE_SIZE);
+    os_alloc_gc_space(IMMOBILE_TEXT_CORE_SPACE_ID, 0, (char*)TEXT_SPACE_START,
+                      text_space_size);
     gc_init_immobile(); // allocate the page tables
     calc_immobile_space_bounds();
     // Read fixedobj space
     int usage = (uword_t)fixedobj_free_pointer - FIXEDOBJ_SPACE_START;
-    checked_read(fd, (char*)FIXEDOBJ_SPACE_START, usage);
+    checked_read("fixedobj", fd, (char*)FIXEDOBJ_SPACE_START, usage);
     // Always read the whole page table regardless of the current space usage
     int total_npages = FIXEDOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
-    checked_read(fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
-    // Read varyobj space
-    usage = (uword_t)varyobj_free_pointer - VARYOBJ_SPACE_START;
-    checked_read(fd, (char*)VARYOBJ_SPACE_START, usage);
-    total_npages = VARYOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
+    checked_read("fixedobj_PTE", fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
+    // Read text space
+    usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
+    char *tlsf_memory_end = (char*)TEXT_SPACE_START + text_space_size;
+    tlsf_mem_start = preamble.tlsf_mem_start;
+    fprintf(stderr, "tlsf_mem_start=%p\n", tlsf_mem_start);
+    int tlsf_memory_size = tlsf_memory_end - (char*)tlsf_mem_start;
+    checked_read("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
+    memcpy(tlsf_memory_end-3*N_WORD_BYTES, preamble.sentinel_block, 3*N_WORD_BYTES);
+    total_npages = text_space_size / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
-    checked_read(fd, varyobj_page_touched_bits, n_bitmap_elts * sizeof (int));
-    checked_read(fd, varyobj_pages, total_npages * sizeof (int));
+    checked_read("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
+    checked_read("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
+    tlsf_control = preamble.tlsf_control_address; // already mapped at a fixed address
+    // TLSF control was mapped in gc_init_immobile()
+    checked_read("TLSF_control", fd, tlsf_control, preamble.tlsf_control_size);
+    int n_tlsf_pages = tlsf_memory_size / IMMOBILE_CARD_BYTES;
+    fprintf(stderr, "%d TLSF pages\n", n_tlsf_pages);
+    tlsf_page_sso = malloc(n_tlsf_pages * sizeof (short));
+    checked_read("TLSF_sso", fd, tlsf_page_sso, n_tlsf_pages * sizeof (short));
     write_protect_immobile_space();
 #endif
     fprintf(stderr, "%d threads:\n", (int)preamble.nthreads);
     int i;
     for(i=0; i<(int)preamble.nthreads; ++i) {
-        struct thread* th = alloc_thread_struct(0, 0);
+        struct thread* th = alloc_thread_struct(0);
         // Push it on the front
         th->prev = 0;
         th->next = threads;
         if (threads) threads->prev = th;
         threads = th;
-        checked_read(fd, &thread_preamble, sizeof thread_preamble);
+        checked_read("thread", fd, &thread_preamble, sizeof thread_preamble);
         uword_t* stackptr = (uword_t*)((char*)th->control_stack_end - thread_preamble.control_stack_nbytes);
         context = contexts[i] = malloc(preamble.sizeof_context);
         nth_interrupt_context(0, th) = context;
         if (thread_preamble.has_context) {
-            checked_read(fd, context, preamble.sizeof_context);
+            checked_read(" ctxt", fd, context, preamble.sizeof_context);
         }
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
         *os_context_sp_addr(context) = (uword_t)stackptr;
-#else
+#ifdef LISP_FEATURE_X86
+        *os_context_register_addr(context,reg_SP) = (os_context_register_t)stackptr;
+#endif
+        gc_assert(*os_context_register_addr(context,reg_SP) == (os_context_register_t)stackptr);
+#elif defined reg_CSP
         *os_context_register_addr(context, reg_CSP) = (uword_t)stackptr;
 #endif
-        checked_read(fd, stackptr, thread_preamble.control_stack_nbytes);
-        checked_read(fd, th->binding_stack_start, thread_preamble.binding_stack_nbytes);
+        checked_read(" stack", fd, stackptr, thread_preamble.control_stack_nbytes);
+        checked_read(" bindings", fd, th->binding_stack_start, thread_preamble.binding_stack_nbytes);
         // Skip over the initial words of the thread structure that was saved
         // in the file, so that binding_stack_start remains as is in the
         // newly allocated structure. The last word is the only one we want to keep.
         int skip = sizeof dummy-N_WORD_BYTES;
-        checked_read(fd, &dummy, skip);
-        checked_read(fd, &th->lisp_thread, preamble.tls_size-skip);
+        checked_read(0, fd, &dummy, skip);
+        checked_read(" TLS", fd, &th->lisp_thread, preamble.tls_size-skip);
         write_TLS(FREE_INTERRUPT_CONTEXT_INDEX, make_fixnum(1), th);
         struct thread_instance* instance = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
-        lispobj name = instance->name;
-        char* cname = (gc_managed_addr_p(name) &&
-                       widetag_of(native_pointer(name)) == SIMPLE_BASE_STRING_WIDETAG)
-                      ? (char*)name+1 : 0;
+        lispobj name = instance->_name;
+        char* cname = gc_managed_addr_p(name) && simple_base_string_p(name)
+                      ? vector_sap(name) : 0;
         fprintf(stderr, "thread @ %p originally %p, %d bind_stk words, %d val_stk words '%s'\n",
                 th, (void*)thread_preamble.address,
                 (int)(thread_preamble.binding_stack_nbytes>>WORD_SHIFT),
@@ -960,7 +1170,7 @@ int load_gc_crashdump(char* pathname)
         putc('\n', stderr);
     }
     char signature[8];
-    checked_read(fd, signature, 8);
+    checked_read("sig", fd, signature, 8);
     gc_assert(!strncmp(signature, "SB.Crash", 8));
     gc_assert(read(fd, signature, 1) == 0);
     close(fd);
@@ -975,7 +1185,7 @@ int main(int argc, char *argv[], char **envp)
         fprintf(stderr, "Usage: ldb crashdump\n");
         return 1;
     }
-    boolean have_hardwired_spaces = os_preinit(argv, envp);
+    bool have_hardwired_spaces = os_preinit(argv, envp);
     allocate_lisp_dynamic_space(have_hardwired_spaces);
     gc_init();
     load_gc_crashdump(argv[1]);
