@@ -13,9 +13,10 @@
  * files for more information.
  */
 
+#include <stdatomic.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 #include "lispregs.h"
 #include "gc.h"
 #include "code.h"
@@ -24,7 +25,7 @@
 #include "pseudo-atomic.h"
 #include "genesis/instance.h"
 #include "genesis/gc-tables.h"
-#include "genesis/list-node.h"
+#include "genesis/split-ordered-list.h"
 #include "thread.h"
 
 /* forward declarations */
@@ -44,18 +45,6 @@ int n_gcs;
  * and only a few rare messages are printed at level 1. */
 int gencgc_verbose = 0;
 
-/* FIXME: At some point enable the various error-checking things below
- * and see what they say. */
-
-/* We hunt for pointers to old-space, when GCing generations >= verify_gen.
- * Set verify_gens to HIGHEST_NORMAL_GENERATION + 2 to disable this kind of
- * check. */
-generation_index_t verify_gens = HIGHEST_NORMAL_GENERATION + 2;
-
-/* Should we do a pre-scan of the heap before it's GCed? */
-int pre_verify_gen_0 = 0; // FIXME: should be named 'pre_verify_gc'
-
-
 /*
  * GC structures and variables
  */
@@ -267,9 +256,8 @@ static generation_index_t gc_alloc_generation;
 /* We use five regions for the current newspace generation. */
 struct alloc_region gc_alloc_region[6];
 
-static page_index_t
-  alloc_start_pages[8], // one for each value of PAGE_TYPE_x
-  max_alloc_start_page; // the largest of any array element
+static _Atomic struct allocator_state alloc_start_states[8]; // one for each value of PAGE_TYPE_x
+static page_index_t max_alloc_start_page; // the largest of any array element
 page_index_t gencgc_alloc_start_page; // initializer for the preceding array
 
 /* Each 'start_page' informs the region-opening logic where it should
@@ -281,47 +269,42 @@ page_index_t gencgc_alloc_start_page; // initializer for the preceding array
  * It's kind of dumb that there is one start_page per type,
  * other than it serves its purpose for picking up where it left off
  * on a partially full page during GC */
-#define RESET_ALLOC_START_PAGES() \
-        alloc_start_pages[0] = gencgc_alloc_start_page; \
-        alloc_start_pages[1] = gencgc_alloc_start_page; \
-        alloc_start_pages[2] = gencgc_alloc_start_page; \
-        alloc_start_pages[3] = gencgc_alloc_start_page; \
-        alloc_start_pages[4] = gencgc_alloc_start_page; \
-        alloc_start_pages[5] = gencgc_alloc_start_page; \
-        alloc_start_pages[6] = gencgc_alloc_start_page; \
-        alloc_start_pages[7] = gencgc_alloc_start_page; \
-        max_alloc_start_page = gencgc_alloc_start_page;
+void reset_alloc_start_pages(bool allow_free_pages) {
+  for (int i = 0; i < 8; i++)
+    alloc_start_states[i] = (struct allocator_state){gencgc_alloc_start_page, allow_free_pages};
+  max_alloc_start_page = gencgc_alloc_start_page;
+}
 
-static page_index_t
+static struct allocator_state
 get_alloc_start_page(unsigned int page_type)
 {
     if (page_type > 7) lose("bad page_type: %d", page_type);
     struct thread* th = get_sb_vm_thread();
-    page_index_t global_start = alloc_start_pages[page_type];
+    struct allocator_state global_start = alloc_start_states[page_type];
     page_index_t hint;
     switch (page_type) {
     case PAGE_TYPE_MIXED:
-        if ((hint = thread_extra_data(th)->mixed_page_hint) > 0 && hint <= global_start) {
+        if ((hint = thread_extra_data(th)->mixed_page_hint) > 0 && hint <= global_start.page) {
             thread_extra_data(th)->mixed_page_hint = - 1;
-            return hint;
+            return (struct allocator_state){hint, global_start.allow_free_pages};
         }
         break;
     case PAGE_TYPE_CONS:
-        if ((hint = thread_extra_data(th)->cons_page_hint) > 0 && hint <= global_start) {
+        if ((hint = thread_extra_data(th)->cons_page_hint) > 0 && hint <= global_start.page) {
             thread_extra_data(th)->cons_page_hint = - 1;
-            return hint;
+            return (struct allocator_state){hint, global_start.allow_free_pages};
         }
         break;
     }
     return global_start;
 }
 
-static inline void
-set_alloc_start_page(unsigned int page_type, page_index_t page)
+static void
+set_alloc_start_page(unsigned int page_type, struct allocator_state state)
 {
     if (page_type > 7) lose("bad page_type: %d", page_type);
-    if (page > max_alloc_start_page) max_alloc_start_page = page;
-    alloc_start_pages[page_type] = page;
+    if (state.page > max_alloc_start_page) max_alloc_start_page = state.page;
+    alloc_start_states[page_type] = state;
 }
 #include "private-cons.inc"
 
@@ -335,7 +318,7 @@ void gc_close_region(struct alloc_region *alloc_region,
  * when it can't fit in the open region.
  * This entry point is only for use within the GC itself. */
 void *collector_alloc_fallback(struct alloc_region* region, sword_t nbytes, int page_type) {
-    page_index_t alloc_start = get_alloc_start_page(page_type);
+    struct allocator_state alloc_start = get_alloc_start_page(page_type);
     void *new_obj;
     if ((uword_t)nbytes >= (GENCGC_PAGE_BYTES / 4 * 3)) {
         uword_t largest_hole;
@@ -343,6 +326,7 @@ void *collector_alloc_fallback(struct alloc_region* region, sword_t nbytes, int
                                                    &alloc_start, page_table_pages, &largest_hole);
         if (new_page == -1) gc_heap_exhausted_error_or_lose(largest_hole, nbytes);
         new_obj = page_address(new_page);
+        set_allocation_bit_mark(new_obj);
     } else {
         ensure_region_closed(region, page_type);
         bool success =
@@ -781,7 +765,7 @@ garbage_collect_generation(generation_index_t generation, int raise,
 
     /* Change to a new space for allocation, resetting the alloc_start_page */
         gc_alloc_generation = new_space;
-        RESET_ALLOC_START_PAGES();
+        reset_alloc_start_pages(false);
 
         /* Don't try to allocate into pseudo-static, when we collect it */
         if (generation == PSEUDO_STATIC_GENERATION)
@@ -942,7 +926,7 @@ garbage_collect_generation(generation_index_t generation, int raise,
     }
 
     mr_collect_garbage(raise);
-    RESET_ALLOC_START_PAGES();
+    reset_alloc_start_pages(false);
     struct generation* g = &generations[generation];
     /* Set the new gc trigger for the GCed generation. */
     g->gc_trigger = g->bytes_allocated + g->bytes_consed_between_gc;
@@ -952,7 +936,8 @@ garbage_collect_generation(generation_index_t generation, int raise,
     // been on pages that got freed by free_oldspace.
     dynspace_codeblob_tree_snapshot = 0;
     if (generation >= verify_gens)
-        verify_heap(cur_thread_approx_stackptr, VERIFY_POST_GC | (generation<<16));
+        hexdump_and_verify_heap(cur_thread_approx_stackptr,
+                                VERIFY_POST_GC | (generation<<1) | raise);
 
     extern int n_unboxed_instances;
     n_unboxed_instances = 0;
@@ -1206,7 +1191,6 @@ collect_garbage(generation_index_t last_gen)
                      && (generation_average_age(gen)
                          > generations[gen].minimum_age_before_gc))));
 
-#ifdef LISP_FEATURE_SOFT_CARD_MARKS
     {
     // Turn sticky cards marks to the regular mark.
     page_index_t page;
@@ -1218,7 +1202,6 @@ collect_garbage(generation_index_t last_gen)
             gcm[card] = (gcm[card] == STICKY_MARK) ? CARD_MARKED : gcm[card];
     }
     }
-#endif
 
     /* Save the high-water mark before updating next_free_page */
     if (next_free_page > high_water_mark)
@@ -1411,11 +1394,11 @@ lisp_alloc(__attribute__((unused)) int flags,
 #endif
 
     ensure_region_closed(region, page_type);
-    int __attribute__((unused)) ret = mutex_acquire(&free_pages_lock);
-    gc_assert(ret);
-    page_index_t alloc_start = get_alloc_start_page(page_type);
+    struct allocator_state alloc_start = get_alloc_start_page(page_type);
     bool largep = nbytes >= LARGE_OBJECT_SIZE && page_type != PAGE_TYPE_CONS;
     if (largep) {
+        int __attribute__((unused)) ret = mutex_acquire(&free_pages_lock);
+        gc_assert(ret);
         uword_t largest_hole;
         page_index_t new_page = try_allocate_large(nbytes, page_type, gc_alloc_generation,
                                                    &alloc_start, page_table_pages, &largest_hole);
@@ -1424,9 +1407,16 @@ lisp_alloc(__attribute__((unused)) int flags,
         ret = mutex_release(&free_pages_lock);
         gc_assert(ret);
         new_obj = page_address(new_page);
+        set_allocation_bit_mark(new_obj);
         memset(new_obj, 0, nbytes);
     } else {
+        /* Try to find a page before acquiring free_pages_lock. */
+        pre_search_for_small_space(nbytes, page_type, &alloc_start, page_table_pages);
+        int __attribute__((unused)) ret = mutex_acquire(&free_pages_lock);
+        gc_assert(ret);
         if (!gc_active_p) small_allocation_count++;
+        /* This search will only re-visit the page found by pre_search_for_small_space
+         * if no one else claimed the page since acquiring free_pages_lock. */
         bool success =
             try_allocate_small_from_pages(nbytes, region, page_type,
                                           gc_alloc_generation,
@@ -1552,33 +1542,6 @@ static bool card_markedp(void* addr)
     return gc_card_mark[addr_to_card_index(addr)] != CARD_UNMARKED;
 }
 
-char *page_card_mark_string(page_index_t page, char *result)
-{
-    long card = addr_to_card_index(page_address(page));
-    if (cardseq_all_marked_nonsticky(card))
-        result[0] = '*', result[1] = 0;
-    else if (!cardseq_any_marked(card))
-        result[0] = '-', result[1] = 0;
-    else {
-        int i;
-        for(i=0; i<CARDS_PER_PAGE; ++i)
-        switch (gc_card_mark[card+i] & MARK_BYTE_MASK) {
-        case CARD_MARKED: result[i] = '*'; break;
-        case CARD_UNMARKED: result[i] = '-'; break;
-#ifdef LISP_FEATURE_SOFT_CARD_MARKS
-        case STICKY_MARK: result[i] = 'S'; break;
-#else
-        case WP_CLEARED_AND_MARKED: result[i] = 'd'; break; // "d" is for dirty
-#endif
-        default: result[i] = '?'; break; // illegal value
-        }
-        result[CARDS_PER_PAGE] = 0;
-    }
-    return result;
-}
-
-#define PRINT_HEADER_ON_FAILURE 2048
-
 // Check a single pointer. Return 1 if we should stop verifying due to too many errors.
 // (Otherwise continue showing errors until then)
 // NOTE: This function can produces false failure indications,
@@ -1588,10 +1551,10 @@ static void note_failure(lispobj thing, lispobj *where, struct verify_state *sta
                          char *str)
 {
     lock();
-    if (state->flags & PRINT_HEADER_ON_FAILURE) {
+    if (state->flags & VERIFY_PRINT_HEADER_ON_FAILURE) {
         if (state->flags & VERIFY_PRE_GC) fprintf(stderr, "pre-GC failure\n");
         if (state->flags & VERIFY_POST_GC) fprintf(stderr, "post-GC failure\n");
-        state->flags &= ~PRINT_HEADER_ON_FAILURE;
+        state->flags &= ~VERIFY_PRINT_HEADER_ON_FAILURE;
     }
     if (state->object_addr) {
         lispobj obj = compute_lispobj(state->object_addr);
@@ -1614,8 +1577,6 @@ static void note_failure(lispobj thing, lispobj *where, struct verify_state *sta
 static int
 verify_pointer(lispobj thing, lispobj *where, struct verify_state *state)
 {
-    extern int valid_lisp_pointer_p(lispobj);
-
     /* Strict containment: no pointer from a heap space may point
      * to anything outside of a heap space. */
     // bool strict_containment = state->flags & VERIFY_FINAL;
@@ -1705,23 +1666,14 @@ verify_pointer(lispobj thing, lispobj *where, struct verify_state *state)
          *    (either/or because Lisp marks the header card,
          *     but the collector marks the cell's card.) */
         int marked = card_markedp(where)
-#ifdef LISP_FEATURE_SOFT_CARD_MARKS
         || (state->object_header
             && header_widetag(state->object_header) != SIMPLE_VECTOR_WIDETAG
-            && card_markedp(state->object_addr))
-#elif defined LISP_FEATURE_EXECUTABLE_FUNINSTANCES
-        /* #+(and (not soft-card-marks) executable-funinstances) could find the mark
-         * for a page-spanning funinstance on the preceding page, because it uses
-         * logical marking, not physical protection of the page holding the pointer */
-        || (header_widetag(state->object_header) == FUNCALLABLE_INSTANCE_WIDETAG
-            && card_markedp(state->object_addr))
-#endif
-        ;
+            && card_markedp(state->object_addr));
         FAIL_IF(!marked, "younger obj from WP page");
     }
     int valid;
     if (state->flags & VERIFY_AGGRESSIVE) // Extreme paranoia mode
-        valid = valid_lisp_pointer_p(thing);
+        valid = valid_tagged_pointer_p(thing);
     else {
         /* Efficiently decide whether 'thing' is plausible.
          * This MUST NOT use properly_tagged_descriptor_p() which
@@ -1778,6 +1730,7 @@ static int verify_headered_object(lispobj* object, sword_t nwords,
         }
 #endif
         for (i=2; i <nheader_words; ++i) CHECK(object[i], object+i);
+#ifndef NDEBUG // avoid "unused" warnings on auto vars of for_each_simple_fun()
         // Check the SIMPLE-FUN headers
         for_each_simple_fun(i, fheaderp, code, 1, {
 #if defined LISP_FEATURE_COMPACT_INSTANCE_HEADER
@@ -1787,6 +1740,7 @@ static int verify_headered_object(lispobj* object, sword_t nwords,
             gc_assert((fheaderp->header >> 32) == 0);
 #endif
         });
+#endif
 #if 0 // this looks redundant. It's checked with each pointer, no?
         bool rememberedp = header_rememberedp(code->header);
         /* The remembered set invariant is that an object is marked "written"
@@ -1837,13 +1791,6 @@ static __attribute__((unused)) bool acceptable_filler_cons_p(lispobj* where)
 static int verify_range(lispobj* start, lispobj* end, struct verify_state* state)
 {
     lispobj* where = start;
-    if (state->flags & VERIFYING_GENERATIONAL && find_page_index(start)>=0) {
-#ifndef LISP_FEATURE_MARK_REGION_GC /* We use a global bitmap */
-        page_index_t page = find_page_index(start);
-        if (page_table[page].type == PAGE_TYPE_CONS)
-            gc_assert(page_words_used(page) <=  MAX_CONSES_PER_PAGE*CONS_SIZE);
-#endif
-    }
     if ((state->flags & VERIFYING_UNFORMATTED)) {
         while (where < end) {
             if (*where != NO_TLS_VALUE_MARKER) {
@@ -1861,6 +1808,11 @@ static int verify_range(lispobj* start, lispobj* end, struct verify_state* state
          * tests before calling the sizer.  Otherwise the lossage message would
          * not be as good as it could be. I guess that failure doesn't happen much */
         sword_t nwords = object_size(where);
+        if (find_page_index(where) != -1)
+          /* Check that no allocation bits are interior to this object  */
+          for (lispobj *interior = where + 2; interior < where + nwords; interior += 2)
+            if (allocation_bit_marked(interior))
+                lose("object @ %p has an allocation marked interior to it @ %p", where, interior);
         state->object_addr = where;
         state->object_header = is_cons_half(*where) ? 0 : *where;
         if (state->flags & VERIFYING_GENERATIONAL) {
@@ -1978,6 +1930,16 @@ static void parallel_walk_generation(uword_t (*proc)(lispobj*,lispobj*,uword_t),
     os_deallocate((char*)span_table, table_size);
 }
 
+static void check_free_pages()
+{
+    unsigned char *allocs = (unsigned char*)allocation_bitmap;
+    for (page_index_t p = 0; p < page_table_pages; p++)
+        if (page_free_p(p))
+            for_lines_in_page(l, p)
+                if (allocs[l])
+                    lose("You call page #%ld free, despite the fact it's obviously got allocation bits.", p);
+}
+
 /* Return the number of verification errors found.
  * You might want to use that as a deciding factor for dump the heap
  * to a file (which takes time, and generally isn't needed).
@@ -1994,11 +1956,12 @@ int verify_heap(__attribute__((unused)) lispobj* cur_thread_approx_stackptr,
     if (verbose)
         fprintf(stderr,
                 flags & VERIFY_PRE_GC ? "Verify before GC" :
-                flags & VERIFY_POST_GC ? "Verify after GC(%d)" :
+                flags & VERIFY_POST_GC ? "Verify after GC(%d,%d)" :
                 "Heap check", // if called at a random time
-                (int)(flags>>16)); // generation number
+                (flags >> 1) & 7, // generation number
+                flags & 1); // 'raise'
     else
-        state.flags |= PRINT_HEADER_ON_FAILURE;
+        state.flags |= VERIFY_PRINT_HEADER_ON_FAILURE;
 
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
 #  ifdef __linux__
@@ -2047,6 +2010,7 @@ int verify_heap(__attribute__((unused)) lispobj* cur_thread_approx_stackptr,
     state.flags |= VERIFYING_GENERATIONAL;
     parallel_walk_generation((uword_t(*)(lispobj*,lispobj*,uword_t))verify_range,
                              -1, (uword_t)&state);
+    check_free_pages();
     if (verbose && state.nerrors==0) fprintf(stderr, " passed\n");
  out:
     if (state.nerrors && !(flags & VERIFY_DONT_LOSE)) {
@@ -2096,7 +2060,63 @@ void gc_show_pte(lispobj obj)
     printf("not in GC'ed space\n");
 }
 
-int hexdump_and_verify_heap(lispobj* cur_thread_approx_stackptr, int flags)
-{
-    return verify_heap(cur_thread_approx_stackptr, flags);
+/* The other implementation of gc_gen_report_to_file gets the generations
+ * of small objects wrong, as it only checks the generations of pages. */
+void gc_gen_report_to_file(int filedes, FILE *file) {
+  uword_t small_generation_type[PSEUDO_STATIC_GENERATION + 1][8] = { 0 };
+  uword_t small_generation_total[8] = { 0 };
+  uword_t large_generation_type[PSEUDO_STATIC_GENERATION + 1][8] = { 0 };
+  uword_t large_generation_total[8] = { 0 };
+  for (page_index_t p = 0; p < page_table_pages; p++) {
+    unsigned char type = page_table[p].type & PAGE_TYPE_MASK;
+    if (page_free_p(p)) continue;
+    else if (page_single_obj_p(p)) {
+      generation_index_t gen = page_table[p].gen;
+      large_generation_type[gen][type] += GENCGC_PAGE_BYTES;
+      large_generation_total[gen] += GENCGC_PAGE_BYTES;
+    } else {
+      for_lines_in_page (l, p) {
+        if (line_bytemap[l]) {
+          generation_index_t gen = DECODE_GEN(line_bytemap[l]);
+          small_generation_type[gen][type] += LINE_SIZE;
+          small_generation_total[gen] += LINE_SIZE;
+        }
+      }
+    }
+  }
+#define OUTPUT(str) \
+  {if (file) fwrite(str, 1, strlen(str), file); if (filedes>=0) ignore_value(write(filedes, str, strlen(str)));}
+  char buffer[180];
+#define FORMAT(control, ...) snprintf(buffer, sizeof(buffer), control, __VA_ARGS__); OUTPUT(buffer)
+#define MB(n) (int)((n) >> 20)
+  FORMAT("  %4s %6s %6s %6s %6s %6s %6s\n",
+         "Gen#", "Raw", "Boxed", "Mixed", "Cons", "Code", "Total");
+  OUTPUT("Small objects (in MiB):\n");
+  for (generation_index_t g = 0; g <= HIGHEST_NORMAL_GENERATION; g++)
+    if (small_generation_total[g]) {
+      uword_t *gen = small_generation_type[g];
+      FORMAT("  %4d %6d %6d %6d %6d %6d %6d\n",
+             g,
+             MB(gen[PAGE_TYPE_UNBOXED]),
+             MB(gen[PAGE_TYPE_BOXED]),
+             MB(gen[PAGE_TYPE_MIXED]),
+             MB(gen[PAGE_TYPE_CONS]),
+             MB(gen[PAGE_TYPE_CODE]),
+             MB(small_generation_total[g]));
+    }
+  OUTPUT("Large objects:\n");
+  for (generation_index_t g = 0; g <= PSEUDO_STATIC_GENERATION; g++)
+    if (large_generation_total[g]) {
+      uword_t *gen = large_generation_type[g];
+      FORMAT("  %4d %6d %6d %6d %6d %6d %6d\n",
+             g,
+             MB(gen[PAGE_TYPE_UNBOXED]),
+             MB(gen[PAGE_TYPE_BOXED]),
+             MB(gen[PAGE_TYPE_MIXED]),
+             MB(gen[PAGE_TYPE_CONS]),
+             MB(gen[PAGE_TYPE_CODE]),
+             MB(large_generation_total[g]));
+    }
+#undef FORMAT
+#undef OUTPUT
 }
